# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import re


def _note_into_doc(note, doc):
    """Insert a note into a docstring under the head line.

    Parameters
    ----------
    note : str
        Note to insert.
    doc : str
        Docstring to modify.

    Returns
    -------
    str
        Modified docstring.

    Notes
    -----
    The note will be placed below the first paragraph (usually a single line) of the
    docstring, like a subtitle.

    """
    # Make sure note has one line end.
    note = note.rstrip() + "\n"

    # No docstring: note becomes the entire docstring.
    if not doc:
        return note

    # Determine the indentation of the second non-empty line.
    # Correct indentation is important for docstring rendering.
    if match := re.search(r"\n+(\s*)\S*", doc):
        indent = match.group(1)

    # If such a line doesn't exist, i.e., the docstring is a single line, then
    # indentation is empty.
    else:
        indent = ""

    # Find the first paragraph break (an empty line) and insert the note after the
    # first paragraph.
    if pos := doc.find("\n\n") + 1:
        return f"{doc[:pos].rstrip()}\n\n{indent}{note}{doc[pos:]}"

    # If not found, i.e., the docstring is a single paragraph, then append the note to
    # the end of docstring.
    else:
        return f"{doc.rstrip()}\n\n{indent}{note}{indent}"


def _note_into_doc_param(note, doc, param):
    """Insert a note into a docstring under the description of a parameter.

    Parameters
    ----------
    note : str
        Note to insert.
    doc : str
        Docstring to modify.
    param : str
        Target parameter.

    Returns
    -------
    str
        Modified docstring.

    Raises
    ------
    ValueError
        Parameter is missing or its format is invalid.

    Notes
    -----
    The note will be placed below the description of the parameter, and above the next
    parameter or the next section.

    """
    note = note.rstrip() + "\n"

    # Find the header line of the parameter.
    if not (match := re.search(rf"(^\s*){param}\s*:.*?\n", doc, re.MULTILINE)):
        raise ValueError(
            f'Parameter "{param}" is missing from the docstring or its format is '
            "invalid."
        )

    # Determine the indentation of the header line.
    indent = match.group(1)

    # Find the next line with the same or less indentation, which indicates the next
    # parameter or the end of the "Parameters" section.
    after = doc[(end := match.end()) :]
    match = re.search(rf"^{indent}(\S|$)", after, re.MULTILINE)

    # Determine the insertion point.
    pos = end + (match.start() if match else len(after))

    # Insert the message. It should have 1+ indentation level (i.e., 4 spaces) than
    # the header line.
    return f"{doc[:pos].rstrip()}\n\n{indent}    {note}\n{doc[pos:]}"


def _deprecation_note(ver=None, msg=None):
    """Create a note indicating deprecation."""
    note = f".. deprecated:: {ver}"
    if msg:
        note += " " + msg
    return note


def _renaming_note(name, ver=None, warn=False):
    """Create a message indicating renaming."""
    if not ver:
        return f"Alias: ``{name}``"
    note = (
        f".. versionchanged:: {ver} "
        f"Renamed from ``{name}``. The old name is kept as an alias"
    )
    if warn:
        note += " but is deprecated"
    note += "."
    return note



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

cimport cython
from libc.float cimport DBL_MAX


@cython.boundscheck(False)
@cython.wraparound(False)
def nj_minq_cy(double[:, :] dm, double[:] sums):
    r"""Find the minimum value in a Q-matrix during the NJ algorithm.

    Parameters
    ----------
    dm : (N, N) ndarray
        Distance matrix.
    sums : (N,) ndarray
        Distance sum vector.

    Returns
    -------
    (int, int)
        Row and column indices of the minimum value.

    See Also
    --------
    skbio.tree.nj

    Notes
    -----
    The agglomerative clustering algorithm neighbor joining (NJ) involves converting
    the original distance matrix into a "Q-matrix" and finding the location of the
    minimum _q_ value in the matrix.

    .. math::

        Q(i, j) = (n - 2) d(i, j) - \sum d(i) - \sum d(j)

    This function calculates _q_ values and finds the minimum as the calculation goes,
    and avoids creating the entire Q-matrix, thereby saving compute.

    """
    cdef Py_ssize_t n = dm.shape[0]
    cdef Py_ssize_t i, j
    cdef int n_2 = n - 2

    # current minimum q value and its location
    cdef Py_ssize_t min_i, min_j
    cdef double min_q = DBL_MAX

    # current q value and minimum q value plus \sum d(i)
    cdef double q_plus, min_q_plus

    # loop the upper-right triangle of the distance matrix
    # i < j is guaranteed
    for i in range(n):
        min_q_plus = min_q + sums[i]
        for j in range(i + 1, n):
            q_plus = dm[i, j] * n_2 - sums[j]
            if q_plus < min_q_plus:
                min_q_plus = q_plus
                min_q = q_plus - sums[i]
                min_i, min_j = i, j

    return min_i, min_j



import io
from unittest import TestCase, main

from skbio import DistanceMatrix, TreeNode

from skbio.tree._upgma import (
    upgma
)


class UpgmaTests(TestCase):

    def setUp(self):
        data = [[0,  5,  9,  9,  8],
                [5,  0, 10, 10,  9],
                [9, 10,  0,  8,  7],
                [9, 10,  8,  0,  3],
                [8,  9,  7,  3,  0]]
        ids = list('abcde')
        self.dm = DistanceMatrix(data, ids)
        # Note that upgma and wpgma have different edge
        # estimations, even with small examples.
        self.expected_str_upgma = (
            "((a:2.5,b:2.5):2.083333333333333,(c:3.75,(d:1.5,e:1.5):2.25):0.833333333333333);")
        self.expected_str_wpgma = (
            "((a:2.5,b:2.5):2.125,(c:3.75,(d:1.5,e:1.5):2.25):0.875);")
        self.expected_TreeNode_upgma = TreeNode.read(
            io.StringIO(self.expected_str_upgma))
        self.expected_TreeNode_wpgma = TreeNode.read(
            io.StringIO(self.expected_str_wpgma))

    def test_upgma_dm(self):
        actual_TreeNode = upgma(self.dm)
        self.assertAlmostEqual(actual_TreeNode.compare_cophenet(
            self.expected_TreeNode_upgma), 0.0, places=10)

    def test_wpgma_dm(self):
        actual_TreeNode = upgma(self.dm, weighted=True)
        self.assertAlmostEqual(actual_TreeNode.compare_cophenet(
            self.expected_TreeNode_wpgma), 0.0, places=10)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from heapq import heapify, heappop

import numpy as np

from skbio.tree import TreeNode
from ._c_me import (
    _preorder,
    _postorder,
    _insert_taxon,
    _avgdist_taxon,
    _bal_avgdist_taxon,
    _avgdist_d2_insert,
    _bal_avgdist_insert,
    _ols_lengths,
    _ols_lengths_d2,
    _bal_lengths,
    _ols_min_branch_d2,
    _bal_min_branch,
    _avgdist_matrix,
    _bal_avgdist_matrix,
    _avgdist_swap,
    _bal_avgdist_swap,
    _ols_all_swaps,
    _ols_corner_swaps,
    _bal_all_swaps,
    _bal_avgdist_insert_p,
)
from ._utils import _check_dm, _check_dm_tree


def gme(dm, neg_as_zero=True):
    r"""Perform greedy minimum evolution (GME) for phylogenetic reconstruction.

    .. versionadded:: 0.6.2

    .. versionchanged:: 0.6.3
        Computational efficiency significantly improved.

    Parameters
    ----------
    dm : skbio.DistanceMatrix
        Input distance matrix containing distances between taxa.
    neg_as_zero : bool, optional
        If True (default), convert negative branch lengths into zeros.

    Returns
    -------
    TreeNode
        Reconstructed phylogenetic tree.

    See Also
    --------
    bme
    nni
    nj

    Notes
    -----
    Greedy Minimum Evolution (GME) [1]_ is a distance-based algorithm for phylogenetic
    reconstruction. It utilizes the minimum evolution (ME) principle [2]_ for selecting
    a tree topology with the lowest sum of branch lengths, as calculated using an
    ordinary least squares (OLS) framework [3]_.

    GME is *O*\(*n*:sup:`2`) in time and *O*\(*n*) in space, making it a more scalable
    method than multiple alternatives (e.g., :func:`nj` and :func:`bme`). Therefore it
    is suitable for reconstructing very large phylogenetic trees.

    GME generates an unrooted tree with variable tip heights and potentially some
    negative branch lengths. The first taxon in the distance matrix is always placed as
    a child of the root node. See the notes of :func:`nj` for how to deal with unrooted
    trees and negative branch lengths.

    A GME-generated tree may be further improved by executing the FastNNI algorithm
    implemented in :func:`nni` (with ``balanced=False``).

    A similar but less scalable algorithm using the balanced instead of OLS framework
    is provided in :func:`bme`.

    The same methods underlying :func:`gme`, :func:`bme` and :func:`nni` were also
    provided by the software package FastME [4]_.

    .. note::
        These scikit-bio functions were implemented following the original paper [1]_.
        It is not guaranteed that they will precisely mirror FastME's output. Although
        in practices they usually generate identical or equally optimal phylogenetic
        trees as FastME does.

    References
    ----------
    .. [1] Desper, R., & Gascuel, O. (2002). Fast and accurate phylogeny reconstruction
       algorithms based on the minimum-evolution principle. J Comput Biol, 9(5),
       687-705.

    .. [2] Rzhetsky, A., & Nei, M. (1993). Theoretical foundation of the
       minimum-evolution method of phylogenetic inference. Mol Biol Evol, 10(5),
       1073-1095.

    .. [3] Cavalli-Sforza, L. L., & Edwards, A. W. (1967). Phylogenetic analysis.
       Models and estimation procedures. American journal of human genetics,
       19(3 Pt 1), 233.

    .. [4] Lefort, V., Desper, R., & Gascuel, O. (2015). FastME 2.0: a comprehensive,
       accurate, and fast distance-based phylogeny inference program. Mol Biol Evol,
       32(10), 2798-2800.

    Examples
    --------
    Define a new distance matrix object describing the distances between five taxa:
    human, monkey, pig, rat, and chicken.

    >>> from skbio import DistanceMatrix
    >>> from skbio.tree import gme

    >>> dm = DistanceMatrix([[0, 0.02,  0.18,  0.34,  0.55],
    ...                      [0.02,  0, 0.19, 0.35,  0.55],
    ...                      [0.18, 0.19,  0,  0.34,  0.54],
    ...                      [0.34, 0.35,  0.34,  0,  0.62],
    ...                      [0.55,  0.55,  0.54,  0.62,  0]],
    ...                     ['human','monkey','pig','rat','chicken'])

    Perform Greedy Minimum Evoltuion (GME) and construct the minimum evolution tree
    representing the relationship between those taxa. This is returned as a TreeNode
    object.

    >>> tree = gme(dm)
    >>> print(tree.ascii_art())
              /-monkey
             |
             |          /-pig
             |---------|
    ---------|         |          /-rat
             |          \--------|
             |                    \-chicken
             |
              \-human

    """
    _check_dm(dm)

    # reconstruct tree topology and branch lengths using GME
    tree, lens = _gme(dm.data)

    if neg_as_zero:
        lens[lens < 0] = 0

    return _to_treenode(tree, dm.ids, lens, unroot=True)


def bme(dm, neg_as_zero=True, **kwargs):
    r"""Perform balanced minimum evolution (BME) for phylogenetic reconstruction.

    .. versionadded:: 0.6.3

    Parameters
    ----------
    dm : skbio.DistanceMatrix
        Input distance matrix containing distances between taxa.
    neg_as_zero : bool, optional
        If True (default), convert negative branch lengths into zeros.

    Returns
    -------
    TreeNode
        Reconstructed phylogenetic tree.

    See Also
    --------
    gme
    nni
    nj

    Notes
    -----
    Balanced Minimum Evolution (BME) [1]_ is a refinement of the distance-based minimum
    evolution problem where the average distances between subtrees are independent of
    the sizes of the subtrees. This is referred to as a balanced (or simply BME)
    framework [2]_, as in contrast to the OLS framework used by GME (:func:`gme`).

    The BME algorithm implemented here uses a similar greedy algorithm as implemented
    in :func:`gme`, but less scalable due to the need to update subtree distances as
    the tree topology changes. The algorithm is sub-*O*\(*n*:sup:`3`) in time and
    *O*\(*n*:sup:`2`) in space.

    Refer to :func:`gme` for the format of the output tree and subsequent treatments.

    A BME-generated tree may be further improved by executing the BNNI algorithm
    implemented in :func:`nni` (with ``balanced=True``).

    The same method was provided by FastME [3]_. See :func:`gme` for notes on this.

    .. note::
        Experimental feature: Add ``parallel=True`` will enable parallelization,
        which may increase the performance of the algorithm. This feature may not
        be stable and may be modified without notice in the future.

    References
    ----------
    .. [1] Desper, R., & Gascuel, O. (2002). Fast and accurate phylogeny reconstruction
       algorithms based on the minimum-evolution principle. J Comput Biol, 9(5),
       687-705.

    .. [2] Pauplin, Y. (2000). Direct calculation of a tree length using a distance
       matrix. J Mol Evol, 51, 41-47.

    .. [3] Lefort, V., Desper, R., & Gascuel, O. (2015). FastME 2.0: a comprehensive,
       accurate, and fast distance-based phylogeny inference program. Mol Biol Evol,
       32(10), 2798-2800.

    Examples
    --------
    Define a new distance matrix object describing the distances between five taxa:
    human, monkey, pig, rat, and chicken.

    >>> from skbio import DistanceMatrix
    >>> from skbio.tree import bme

    >>> dm = DistanceMatrix([[0, 0.02,  0.18,  0.34,  0.55],
    ...                      [0.02,  0, 0.19, 0.35,  0.55],
    ...                      [0.18, 0.19,  0,  0.34,  0.54],
    ...                      [0.34, 0.35,  0.34,  0,  0.62],
    ...                      [0.55,  0.55,  0.54,  0.62,  0]],
    ...                     ['human','monkey','pig','rat','chicken'])

    Perform Balanced Minimum Evoltuion (BME) and construct the minimum evolution tree
    representing the relationship between those taxa. This is returned as a TreeNode
    object.

    >>> tree = bme(dm)
    >>> print(tree.ascii_art())
              /-monkey
             |
             |          /-pig
             |---------|
    ---------|         |          /-rat
             |          \--------|
             |                    \-chicken
             |
              \-human

    """
    _check_dm(dm)

    # reconstruct tree topology and branch lengths using BME
    tree, lens = _bme(dm.data, **kwargs)

    if neg_as_zero:
        lens[lens < 0] = 0

    return _to_treenode(tree, dm.ids, lens, unroot=True)


def nni(tree, dm, balanced=True, neg_as_zero=True):
    r"""Perform nearest neighbor interchange (NNI) to improve a phylogenetic tree.

    .. versionadded:: 0.6.2

    .. versionchanged:: 0.6.3
        Computational efficiency significantly improved.

    Parameters
    ----------
    tree : skbio.TreeNode
        Input phylogenetic tree to be rearranged.
    dm : skbio.DistanceMatrix
        Input distance matrix containing distances between taxa.
    balanced : bool, optional
        Use the OLS framework (False) or the balanced framework (True, default).
    neg_as_zero : bool, optional
        If True (default), convert negative branch lengths into zeros.

    Returns
    -------
    TreeNode
        Rearranged phylogenetic tree.

    Notes
    -----
    Nearest neighbor interchange (NNI) is a method for tree rearrangement aiming at
    optimizing a given tree topology according to certain criteria. It iteratively
    swaps neighboring branches that could result in a better tree, until no such swaps
    remain in the optimized tree.

    This function performs NNI for the minimum evolution (ME) problem on phylogenetic
    trees. The implementation is based on [1]_. Two versions of the method are
    provided: the FastNNI algorithm (``balanced=False``) based on an ordinary least
    squares (OLS) framework (see also :func:`gme`) and the BNNI algorithm
    (``balanced=True``) based on a balanced framework (see also :func:`bme`). The
    former is more scalable than the latter.

    The input tree may be rooted or unrooted, and it must be strictly bifurcating. One
    can apply :meth:`~TreeNode.is_bifurcating` or :meth:`~TreeNode.bifurcate` to check
    or make a bifurcating tree. The set of taxa in the tree (accessible via
    :meth:`~TreeNode.subset`) must match those in the distance matrix.

    The output tree is unrooted, with the first taxon in the distance matrix placed as
    a child of the root node. Branch lengths are assigned according to the framework of
    choice. See also :func:`nj` regarding rooting an unrooted tree and dealing with
    potential negative branch lengths.

    The same methods were provided by FastME [2]_. See :func:`gme` for notes on this.

    References
    ----------
    .. [1] Desper, R., & Gascuel, O. (2002). Fast and accurate phylogeny reconstruction
       algorithms based on the minimum-evolution principle. J Comput Biol, 9(5),
       687-705.

    .. [2] Lefort, V., Desper, R., & Gascuel, O. (2015). FastME 2.0: a comprehensive,
       accurate, and fast distance-based phylogeny inference program. Mol Biol Evol,
       32(10), 2798-2800.

    Examples
    --------
    >>> from skbio import DistanceMatrix, TreeNode
    >>> from skbio.tree import nni

    Define a distance matrix describing the distances between five taxa.

    >>> dm = DistanceMatrix([[0, 0.02,  0.18,  0.34,  0.55],
    ...                      [0.02,  0, 0.19, 0.35,  0.55],
    ...                      [0.18, 0.19,  0,  0.34,  0.54],
    ...                      [0.34, 0.35,  0.34,  0,  0.62],
    ...                      [0.55,  0.55,  0.54,  0.62,  0]],
    ...                     ['human','monkey','pig','rat','chicken'])

    Provide a tree to be rearranged. The tree must be bifurcating. Branches lengths are
    not required.

    >>> tree = TreeNode.read(["((human,chicken),(rat,monkey),pig);"])
    >>> print(tree.ascii_art())
                        /-human
              /--------|
             |          \-chicken
             |
    ---------|          /-rat
             |---------|
             |          \-monkey
             |
              \-pig

    Perform nearest neighbor interchange (NNI). This will rearrange tree topolgy to
    better approach the minimum evolution criterion.

    >>> tree = nni(tree, dm)
    >>> print(tree.ascii_art())
                        /-pig
              /--------|
             |         |          /-chicken
             |          \--------|
    ---------|                    \-rat
             |
             |--monkey
             |
              \-human

    Besides rearranging tree topology, estimated branch lengths are assigned to the
    tree.

    >>> rat = tree.find('rat')
    >>> print(rat.length)
    0.20875

    """
    _check_dm_tree(dm, tree)

    # generate tree array
    taxa = dm.ids
    tree, preodr, postodr = _root_from_treenode(tree, taxa)

    # allocate lengths
    lens = np.empty(len(tree), dtype=float)

    # perform BNNI or FastNNI
    func = _bnni if balanced else _fastnni
    func(dm.data, tree, preodr, postodr, lens)

    if neg_as_zero:
        lens[lens < 0] = 0

    # generate TreeNode object
    return _to_treenode(tree, taxa, lens, unroot=True)


def _gme(dm):
    r"""Perform greedy minimum evolution (GME) for phylogenetic reconstruction.

    Parameters
    ----------
    dm : ndarray of float of shape (m, m)
        Input distance matrix containing distances between taxa.

    Returns
    -------
    tree : ndarray of int of shape (2m - 3, 4)
        Reconstructed tree topology by GME.
    lens : ndarray of float of shape (2m - 3,)
        Estimated branch lengths by OLS.

    Notes
    -----
    The GME algorithm involves growing an unrooted tree by iteratively inserting taxa.
    In the beginning it initializes a triplet tree with the first three taxa:

          0
          |
          x
         / \
        1   2

    Taxon 0 is placed at the root position, with its unique descendant x serving as
    connector for taxa 1 and 2 as its left and right children. Next, taxon 3 will
    be inserted into a branch that suffices the minimum evolution criterion. Let's say
    the branch connecting x and 2. The tree becomes:

          0
          |
          x
         / \
        1   y
           / \
          2   3

    So on so forth until all taxa are included in the tree.

    Finally, branch lengths are estimated using an ordinary least squares (OLS)
    framework.

    ---

    This implementation uses an array-based tree structure to improve efficiency. See
    :func:`_check_tree` for details of this structure. Basically, the root node (taxon
    0) is omitted, making the tree strictly binary. Instead, the unique descendant (x)
    is now treated as the root and marked with taxon 0:

          0
         / \
        1   y
           / \
          2   3

    The root is always the first row in the tree array. Other nodes are appended to the
    tree array in the same order as they are added to the tree.

    ---

    For an input distance matrix with m taxa, the output tree will have 2m - 3 nodes.
    This memory space is pre-allocated to improve efficiency. The algorithm doesn't
    remake arrays during iteration, but repeatedly uses the already allocated space.
    During an interation involving x nodes, the first x positions of each array are
    occupied, while the remaining positions are disregarded. There is no need to reset
    array values after each iteration.

    Since they are created de novo, all arrays are C-continuous. This permits further
    optimization in the Cython code.

    """
    # number of taxa
    m = dm.shape[0]

    # number of nodes in the final tree
    n = 2 * m - 3

    # Pre-allocate memory spaces:
    # tree structure
    tree, preodr, postodr = _allocate_tree(n)

    # average distances between distant-2 subtrees
    ad2 = np.empty((n, 2), dtype=float)

    # average distances from a taxon to each subtree
    adk = np.empty((n, 2), dtype=float)

    # branch lengths or length changes
    lens = np.empty((n,), dtype=float)

    # Initialize 3-taxon tree.
    _init_tree(dm, tree, preodr, postodr, ad2, matrix=False)

    # Iteratively add taxa to the tree.
    for k in range(3, m):
        # Calculate average distances from new taxon to existing subtrees.
        _avgdist_taxon(adk, k, dm, tree, preodr, postodr)

        # Find a branch with minimum length change.
        target = _ols_min_branch_d2(lens, ad2, adk, tree, preodr)

        # Update average distances between distant-2 subtrees.
        _avgdist_d2_insert(ad2, target, adk, tree, preodr)

        # Insert new taxon into tree.
        _insert_taxon(k, target, tree, preodr, postodr, use_depth=False)

    # Calculate branch lengths using an OLS framework.
    _ols_lengths_d2(lens, ad2, tree)

    return tree, lens


def _bme(dm, parallel=False):
    r"""Perform balanced minimum evolution (BME) for phylogenetic reconstruction.

    Parameters
    ----------
    dm : ndarray of float of shape (m, m)
        Input distance matrix containing distances between taxa.

    Returns
    -------
    tree : ndarray of int of shape (2m - 3, 4)
        Reconstructed tree topology by BME.
    lens : ndarray of float of shape (2m - 3,)
        Estimated branch lengths by a balanced framework.

    Notes
    -----
    The BME algorithm resembles GME but uses a balanced instead of OLS framework.
    See :func:`_gme` for details of the latter.

    The main differences are: 1) BME requires a full matrix stored and iteratively
    updated, which is less efficient than GME in time and space. 2) BME doesn't rely
    on subtree sizes, which is simpler and more efficient than GME in individual
    calculations (but this cannot compensate for the former).

    """
    func = _bal_avgdist_insert_p if parallel else _bal_avgdist_insert

    # numbers of taxa and nodes in the tree
    m = dm.shape[0]
    n = 2 * m - 3

    # Pre-allocate memory spaces:
    # tree structure
    tree, preodr, postodr = _allocate_tree(n)

    # average distances between all subtrees
    # (This is the dominant factor in BME, and what makes it more expensive than GME.)
    adm = np.empty((n, n), dtype=float)

    # average distances from a taxon to each subtree
    adk = np.empty((n, 2), dtype=float)

    # branch lengths or length changes
    lens = np.empty((n,), dtype=float)

    # a stack for traversal operations
    stack = np.empty((n,), dtype=int)

    # initialize 3-taxon tree
    _init_tree(dm, tree, preodr, postodr, adm, matrix=True)

    # Pre-calculate negative powers of 2.
    powers = np.ldexp(1.0, -np.arange(m))

    # Iteratively add taxa to the tree.
    for k in range(3, m):
        # Calculate balanced average distances from new taxon to existing subtrees.
        _bal_avgdist_taxon(adk, k, dm, tree, preodr, postodr)

        # Find the branch with minimum length change.
        target = _bal_min_branch(lens, adm, adk, tree, preodr)

        # Update balanced average distance matrix between all subtrees.
        func(adm, target, adk, tree, preodr, postodr, powers, stack)

        # Insert new taxon into tree.
        _insert_taxon(k, target, tree, preodr, postodr, use_depth=True)

    # Calculate branch lengths using a balanced framework.
    _bal_lengths(lens, adm, tree)

    return tree, lens


def _fastnni(dm, tree, preodr, postodr, lens):
    r"""Perform fast nearest neighbor interchange (FastNNI) on a tree.

    To improve the tree under the minimum evolution (ME) criterion using an OLS
    framework on a distance matrix between taxa.

    This algorithm was implemented following Section 2.2 and Appendix 4 of Desper and
    Gascuel (2002). Basically, it creates a full average distance matrix between all
    pairs of subtrees, evaluates all possible swaps, then it iteratively perform
    swaps that lead to the maximum reduction of overall branch lengths. During this
    process, it needs to update the evaluation of the four corner branches after one
    branch is swapped.

    ---

    The swaps that can reduce overall branch length are stored in a heap. The paper
    suggests using a maxheap, but the following code uses Python's heapq which is a
    minheap. Therefore the length changes are negated (in contrast to `_bnni`). Each
    item in the heap is a tuple consisting of:

        0. Length change (negative; smaller is better)
        1. Corresponding node index
        2. Corresponding side (0: left, 1: right) child to be swapped with sibling

    For example, (-0.15, 7, 1) means that swapping the right child and the sibling of
    node 7 will reduce the overall branch length by 0.15.

    Meanwhile, the length change of each node is stored in `lens`. For root and tips,
    this value is 0. For internal nodes, this value is negative or 0. Only negative
    values are stored in the heap. 0 means this node is not useful.

    Additionally, the "side" information is stored in column 7 of the tree array.

    When a swap is updated, the new length change (if negative) is pushed into the
    heap, but the old length change won't be deleted from the heap (which would be
    inefficient as O(n)). Instead, a "lazy deletion" mechanism is adopted: When the
    negative-most swap is popped out of the heap, it is checked against the stored
    length change (in `lens`) and side (in `tree[:, 7]`). If they don't match, then
    it means that it is outdated and should be skipped. Otherwise, it is considered
    relevant and the swap will be performed.

    TODO: heapq is a Python module and may not be efficient. Tests showed that the
    code isn't significantly faster than a naive `np.argmax` on then entire `lens`.
    Consider optimization.

    """
    n = tree.shape[0]
    stack = np.empty(n, dtype=int)

    # Calculate average distances between all pairs of subtrees.
    adm = np.empty((n, n))
    _avgdist_matrix(adm, dm, tree, preodr, postodr)

    # Calculate length changes of all possible swaps.
    _ols_all_swaps(lens, tree, adm)

    # Create a heap that stores negative length changes and their nodes and sides.
    heap = [(lens[i], i, tree[i, 7]) for i in np.nonzero(lens)[0]]
    heapify(heap)

    # Iteratively swap branches until there is no more beneficial swap.
    while heap:
        # Pop the swap that reduces the most length.
        L, target, side = heappop(heap)

        # Skip if this swap is outdated.
        if L != lens[target] or side != tree[target, 7]:
            continue

        # Reset this swap.
        lens[target] = 0

        # Swap the branches in the tree.
        _swap_branches(target, side, tree, preodr, stack, use_depth=False)

        # Update average distances after swapping.
        _avgdist_swap(adm, target, side, tree)

        # Update length reductions of swaps of the four corner branches.
        _ols_corner_swaps(target, heap, lens, tree, adm)

    # Calculate branch lengths using an OLS framework.
    _ols_lengths(lens, adm, tree)


def _bnni(dm, tree, preodr, postodr, lens):
    r"""Perform balanced nearest neighbor interchange (BNNI) on a tree.

    To improve the tree under the balanced minimum evolution (BME) criterion given a
    distance matrix between taxa.

    Implemented according to Section 3.2 and Appendix 5 of Desper and Gascuel (2002).
    Basically, BNNI is similar to FastNNI (see `_fastnni`), but it needs to update
    more than the four corner branches after each swap, as balanced average distances
    are dependent on the topology of the tree.

    ---

    The original paper doesn't explain how to store and update positive swaps. Because
    a large proportion of swaps need to be re-evaluated during every iteration, one
    would expect that a heap structure as suggested for the FastNNI algorithm may not
    be efficient. The current code simply stores all swaps (positive or not) and runs
    `np.argmax` on all of them during each iteration.

    """
    n = tree.shape[0]
    stack = np.empty(n, dtype=int)

    # Calculate balanced average distances between all pairs of subtrees.
    adm = np.empty((n, n))
    _bal_avgdist_matrix(adm, dm, tree, preodr, postodr)

    # Pre-calculate negative powers of 2.
    powers = np.ldexp(1.0, -np.arange(dm.shape[0]))

    # Initialize branch swapping information.
    gains, sides, nodes = _init_swaps(tree)

    # Iteratively swap branches until there is no more beneficial swap.
    while True:
        # Calculate length reductions of all possible swaps.
        _bal_all_swaps(gains, sides, nodes, adm, tree)

        # Find the swap with the maximum length reduction, and stop if non-positive.
        branch = gains.argmax()
        if (gain := gains[branch]) <= 0:
            break
        side = sides[branch]
        target = nodes[branch]

        # Swap the branches in the tree.
        _swap_branches(target, side, tree, preodr, stack, use_depth=True)

        # Update balanced average distances after swapping.
        _bal_avgdist_swap(adm, target, side, tree, preodr, powers, stack)

    # Calculate branch lengths using a balanced framework.
    _bal_lengths(lens, adm, tree)


def _check_tree(tree, preodr, postodr):
    r"""Check the integrity of an array-based tree structure.

    The greedy algorithms implemented in this module use a special array-based tree
    structure to improve efficiency.

    In this 2-D array, rows represent nodes in the order of addition to the tree,
    with the root placed at row 0. There are eight columns:

        0. Left child index, or 0 if a tip.
        1. Right child index, or taxon index if a tip.
        2. Parent index, or 0 if root.
        3. Sibling index, or 0 if root.
        4. Size, i.e., number of taxa descending from the node, or 1 if a tip.
        5. Depth, i.e., number of branches constituting the path to root.
        6. Preorder index.
        7. Postorder index.

    Meanwhile, node indices in preorder and postorder are stored in two separate 1-D
    arrays to facilitate traversals.

    Note: Columns 2-7 and the two separate arrays can be derived from columns 0 and 1.
    They are explicitly stored because of their frequent usage during the calculation.
    Also, they are updated iteratively as the tree grows, which is more efficient than
    de novo calculation on an existing tree.

    ---

    For example, a tree with four taxa is like (see :func:`_gme`):

          0
         / \
        1   y
           / \
          2   3

    This tree can be represented by the following 2-D array:

        [[1, 2, 0, 0, 3, 0, 0, 4],
         [0, 1, 0, 2, 1, 1, 1, 0],
         [3, 4, 0, 1, 2, 1, 2, 3],
         [0, 2, 2, 4, 1, 2, 3, 1],
         [0, 3, 2, 3, 1, 2, 4, 2]]

    The separate pre- and postorder arrays are:

        [0, 1, 2, 3, 4]
        [1, 3, 4, 2, 0]

    ---

    This function ensures that a tree array is valid. It is for test purpose, but not
    used in the actual greedy algorithms.

    """
    assert tree.shape[1] == 8
    n = tree[0, 4] * 2 - 1
    assert (tree[:n, 6].argsort() == preodr[:n]).all()
    assert (tree[:n, 7].argsort() == postodr[:n]).all()

    for i in range(n):
        left, right, parent, sibling, size, depth, preidx, postidx = tree[i]
        if left == 0:
            assert i == 0 or right != 0
        else:
            assert tree[left, 2] == tree[right, 2] == i
            assert tree[left, 3] == right
            assert tree[right, 3] == left

    sizes = np.zeros((n,), dtype=int)
    for i in range(n):
        node = postodr[i]
        if tree[node, 0] == 0:
            sizes[node] = 1
        else:
            sizes[node] = sizes[tree[node, :2]].sum()
    assert (sizes == tree[:n, 4]).all()

    depths = np.zeros((n,), dtype=int)
    for i in range(1, n):
        node = preodr[i]
        depths[node] = depths[tree[node, 2]] + 1
    assert (depths == tree[:n, 5]).all()

    stack = np.zeros((n,), dtype=int)
    _preorder(exp := np.zeros((n,), dtype=int), tree, stack)
    assert (exp == preodr[:n]).all()
    _postorder(exp := np.zeros((n,), dtype=int), tree, stack)
    assert (exp == postodr[:n]).all()


def _allocate_tree(n):
    r"""Pre-allocate memory space for an array-based tree structure.

    Parameters
    ----------
    n : int
        Total number of nodes in the tree.

    Returns
    -------
    (n, 8) ndarray of int
        Tree structure.
    (n,) ndarray of int
        Nodes in preorder.
    (n,) ndarray of int
        Nodes in postorder.

    See Also
    --------
    _check_tree

    """
    tree = np.empty((n, 8), dtype=int)
    preodr = np.empty((n,), dtype=int)
    postodr = np.empty((n,), dtype=int)
    return tree, preodr, postodr


def _to_treenode(tree, taxa, lens=None, unroot=False):
    r"""Convert an array-based tree structure into a TreeNode object.

    Parameters
    ----------
    tree : (n, 2+) array_like of int
        Tree structure.
    taxa : list of str
        Taxon names in order.
    lens : (n,) array_like of float, optional
        Branch lengths.
    unroot : bool, optional
        If True, generate an unrooted tree in which the root node has three children,
        the last of which is the first taxon. Otherwise leave the first taxon at the
        root node.

    Returns
    -------
    TreeNode
        Converted TreeNode object.

    See Also
    --------
    TreeNode

    Notes
    -----
    This function only needs the first two columns (left child, right child / taxon
    name) of a tree array to generate a TreeNode object.

    """
    nodes = [TreeNode(taxa[0])] + [
        TreeNode(None if a else taxa[b]) for a, b in tree[1:, :2]
    ]
    if lens is not None:
        for node, L in zip(nodes, lens):
            node.length = L
    for node, (left, right, *_) in zip(nodes, tree):
        if left:
            node.extend([nodes[left], nodes[right]], uncache=False)
    tree = nodes[0]
    if unroot:
        tree.append(TreeNode(taxa[0], length=tree.length), uncache=False)
        tree.name = None
        tree.length = None
    return tree


def _from_treenode(obj, taxmap, tree, preodr, postodr, pos=0, depth=0):
    r"""Convert a TreeNode object into an array-based tree structure.

    Parameters
    ----------
    tree : TreeNode
        Input tree. It must be strictly bifurcating.
    taxmap : dict of str : int
        Mapping of taxon names to indices.
    tree : (n, 8) ndarray of int
        Tree structure.
    preodr : (n,) ndarray of int
        Nodes in preorder.
    postodr : (n,) ndarray of int
        Nodes in postorder.
    pos : int, optional
        Position (axis 0) in the tree array for new data.
    depth : int, optional
        Depth of the root node.

    Raises
    ------
    ValueError
        If tree is not strictly bifurcating.

    See Also
    --------
    _allocate_tree
    _root_from_treenode

    Notes
    -----
    This function fills in pre-allocated tree arrays. It doesn't return any value.

    Nodes in the resulting tree array are ordered by preorder traversal. This may not
    agree with the order they were added to the original tree, if that ever happened.
    Therefore, one shouldn't directly compare the original and generated tree arrays.
    However, the topologies represented by the two tree arrays should be identical.

    The function doesn't fill branch lengths, which aren't required by the downstream
    algorithms. Adding this functionality should be straightforward.

    Parameters ``pos`` and ``depth`` are typically set by :func:`_root_from_treenode`.
    One doesn't need to specify them if working with this function alone. ``depth``
    also impacts the position in ``postodr``. See ``_root_from_treenode`` for
    rationales.

    """
    # determine position in postorder
    post_pos = pos - depth

    # create root node
    tree[pos, 2] = 0
    tree[pos, 3] = 0
    tree[pos, 5] = depth

    # if root node is a tip, just wrap up and return
    if not obj.children:
        tree[pos, 0] = 0
        tree[pos, 1] = taxmap[obj.name]
        tree[pos, 4] = 1
        preodr[pos] = tree[pos, 6] = pos
        postodr[post_pos] = pos
        tree[pos, 7] = post_pos
        return

    # perform preorder traversal, index nodes, fill parent (2), depth (5), taxon (1)
    # and length
    obj.i = pos
    pos_1 = pos + 1
    for i, node in enumerate(obj.preorder(include_self=False)):
        node.i = (ni := pos_1 + i)
        tree[ni, 2] = (pi := node.parent.i)
        tree[ni, 5] = tree[pi, 5] + 1
        if not node.children:
            tree[ni, 0] = 0
            tree[ni, 1] = taxmap[node.name]

    # number of nodes in the current tree
    n = i + 2

    # fill preorder and indices
    end = pos + n
    preodr[pos:end] = tree[pos:end, 6] = np.arange(pos, end)

    # perform postorder traversal, fill children (0 and 1), sibling (3), and size (4)
    for i, node in enumerate(obj.postorder()):
        postodr[post_pos + i] = (ni := node.i)
        tree[ni, 7] = post_pos + i
        if not node.children:
            tree[ni, 4] = 1
        else:
            try:
                left, right = node.children
            except ValueError:
                raise ValueError("Tree is not strictly bifurcating.")
            li, ri = left.i, right.i
            tree[ni, 0] = tree[ri, 3] = li
            tree[ni, 1] = tree[li, 3] = ri
            tree[ni, 4] = tree[li, 4] + tree[ri, 4]
            del left.i
            del right.i

    del obj.i


def _root_from_treenode(obj, taxa):
    r"""Convert a TreeNode object into a tree array rooted at the first taxon.

    Parameters
    ----------
    tree : TreeNode
        Input tree. It must be strictly bifurcating.
    taxa : list of str
        Taxon names in order.
    tree : (n, 8) ndarray of int
        Tree structure.

    Returns
    -------
    (n, 8) ndarray of int
        Tree structure.
    (n,) ndarray of int
        Nodes in preorder.
    (n,) ndarray of int
        Nodes in postorder.

    Raises
    ------
    ValueError
        If tree is not strictly bifurcating.

    See Also
    --------
    _from_treenode
    TreeNode.unrooted_move

    Notes
    -----
    This function locates the first taxon in the tree, walks from it to the root, and
    rotate nodes along the path such that the first taxon becomes the new root in the
    generated tree array. For example (p: parent, c: clade):

          root           taxon
         /  | \           /  \
        c3 p2 c4         c1  p2
          /  \      =>      /  \
         c2  p1           c2  root
            /  \              /  \
        taxon  c1            c3   c4

    Note that the original parent node always becomes the right child of the current
    node. The first parent (p1) is gone and replaced by the taxon. The output array is
    always preordered.

    """
    errmsg = "Tree is not strictly bifurcating."
    taxmap = {taxon: i for i, taxon in enumerate(taxa)}

    m = len(taxa)  # number of taxa
    n = m * 2 - 3  # number of nodes

    # allocate tree arrays
    tree, preodr, postodr = _allocate_tree(n)

    # position of the current node
    pos = 0

    # depth of the current node
    depth = 0

    # position of the previous node
    prev_pos = 0

    # position of the other child of the previous node in the path;
    # it will become the sibling of the current node
    sib_pos = 0

    # number of tips under the current node
    size = m - 1

    n_1 = n - 1

    # helper to add the current node to the array
    def _add_current():
        tree[pos, 2] = prev_pos
        tree[pos, 4] = size
        tree[pos, 5] = depth
        tree[pos, 6] = pos
        preodr[pos] = pos
        tree[pos, 7] = (post_pos := n_1 - depth)
        postodr[post_pos] = pos
        tree[prev_pos, 1] = pos
        tree[sib_pos, 3] = pos
        tree[pos, 3] = sib_pos

    # iterate over ancestors of the first taxon
    curr = obj.find(taxa[0])
    parent = None
    while True:
        prev = curr
        curr = curr.parent
        parent = curr.parent
        siblings = [x for x in curr.children if x is not prev]

        if not siblings:  # singleton branch
            raise ValueError(errmsg)
        n_sibs = len(siblings)
        if n_sibs == 1:
            other = siblings[0]

            # Current node is a regular internal node of the tree (the typical case):
            # Add the current node, then add the other child of it.
            #      parent            prev
            #        |                 |
            #      curr     =>       curr
            #     /   \             /   \
            # other   prev      other   parent
            if parent is not None:
                _add_current()

                prev_pos = pos
                pos += 1

                # add the other child clade
                _from_treenode(other, taxmap, tree, preodr, postodr, pos, depth + 1)

                tree[pos - 1, 0] = pos  # curr's left child
                tree[pos, 2] = pos - 1  # parent (now missing sibling)

                sib_pos = pos
                size -= (k := tree[pos, 4])
                pos += k * 2 - 1

            # Current node is the root of a rooted tree (i.e., it has two children):
            # Don't add the current node, instead add the other child.
            #      curr             prev
            #     /   \     =>      /
            # other   prev      other
            else:
                _from_treenode(other, taxmap, tree, preodr, postodr, pos, depth)

                tree[pos, 2] = prev_pos
                tree[prev_pos, 1] = pos
                tree[sib_pos, 3] = pos
                tree[pos, 3] = sib_pos

                break

        elif n_sibs == 2:
            # Current node is the root of an unrooted tree (i.e., it has three
            # children): Add the current node, then add the other two children of it.
            #                         prev
            #       curr               |
            #     /  |  \    =>       curr
            # left right prev        /   \
            #                     left   right
            if parent is None:
                _add_current()
                curr_pos = pos
                pos += 1

                left, right = siblings
                _from_treenode(left, taxmap, tree, preodr, postodr, pos, depth + 1)
                tree[pos, 2] = curr_pos
                left_pos = pos
                tree[curr_pos, 0] = pos
                pos += tree[pos, 4] * 2 - 1

                _from_treenode(right, taxmap, tree, preodr, postodr, pos, depth + 1)
                tree[pos, 2] = curr_pos
                tree[curr_pos, 1] = pos
                tree[left_pos, 3] = pos
                tree[pos, 3] = left_pos

                break

            else:
                raise ValueError(errmsg)
        else:
            raise ValueError(errmsg)

        # move up one level
        depth += 1

    return tree, preodr, postodr


def _init_tree(dm, tree, preodr, postodr, ads, matrix=False):
    """Initialize tree (triplet with taxa 0, 1 and 2)."""
    # triplet tree
    tree[:3] = [
        [1, 2, 0, 0, 2, 0, 0, 2],  # 0: root
        [0, 1, 0, 2, 1, 1, 1, 0],  # 1: left child
        [0, 2, 0, 1, 1, 1, 2, 1],  # 2: right child
    ]

    # nodes in pre- and postorder
    preodr[:3] = [0, 1, 2]
    postodr[:3] = [1, 2, 0]

    # average distance matrix between all subtrees
    if matrix:
        ads[0, 1] = ads[1, 0] = dm[0, 1]
        ads[0, 2] = ads[2, 0] = dm[0, 2]
        ads[1, 2] = ads[2, 1] = dm[1, 2]

    # average distances between distant-2 subtrees
    else:
        ads[1, 0] = dm[0, 1]
        ads[2, 0] = dm[0, 2]
        ads[1, 1] = ads[2, 1] = dm[1, 2]
        ads[0, 0] = ads[0, 1] = 0


def _insert_taxon_treenode(taxon, target, tree):
    r"""Insert a taxon between a target node and its parent.

    This function resembles :func:`_insert_taxon` but operates on a TreeNode object. It
    is for test purpose. It is not used in the actual algorithms.

    """
    link = TreeNode()
    tip = TreeNode(taxon)
    parent = target.parent
    if parent is None:  # root branch
        link.extend(tree.children)
        tree.append(link)
        tree.append(tip)
    else:  # other branch
        is_left = target is parent.children[0]
        link.append(target)
        link.append(tip)
        parent.append(link)
        if is_left:
            parent.children = parent.children[::-1]


def _avgdist_matrix_naive(adm, dm, tree, postodr):
    r"""Calculate a matrix of average distances between all pairs of subtrees.

    This function produces the same result as :func:`_avgdist_matrix`. However, it
    calculates all subtree-to-subtree distances based on the original taxon-to-taxon
    distances, as discussed in Eq. 1 of Desper and Gascuel (2002). instead of adopting
    a recursive strategy (Eq. 2). Therefore, it is significantly slower.

        d(A, B) = \frac{1}{|A||B|} \sum_{i \in A, j \in B} d(i, j)

    This algorithm is implemented as a reference and for comparison purpose. It is not
    used in the actual GME algorithm.

    """
    n = tree[0, 4] * 2 - 1
    taxas = {}
    for node in postodr[:n]:
        left, right = tree[node, :2]
        if not left:
            taxas[node] = frozenset([right])
        else:
            taxas[node] = taxas[left] | taxas[right]

    full = taxas[0] | frozenset([0])

    for a in range(n):
        a_taxa = taxas[a]
        a_taxa_r = full - a_taxa
        # adm[a, a] = dm[list(a_taxa)][:, list(a_taxa_r)].mean()  # self distance
        for b in range(a + 1, n):
            # If a is an ancestor (proper superset) of b, flip a's taxon set as the
            # the upper subtree of a will be considered.
            a_taxa_ = a_taxa_r if a_taxa > (b_taxa := taxas[b]) else a_taxa
            adm[a, b] = adm[b, a] = dm[list(a_taxa_)][:, list(b_taxa)].mean()
            # Because this is postorder traversal, a cannot be a descendant (proper
            # subset) of b.


def _avgdist_taxon_naive(adk, taxon, dm, tree, postodr):
    """Calculate average distances between a new taxon and existing subtrees.

    This function produces the same result as :func:`_avgdist_taxon`, but it
    calculates all taxon-to-subtree distances based on the original taxon-to-taxon
    distances, as discussed in Eq. 1 of Desper and Gascuel (2002), instead of
    adopting a recursive strategy (Eq. 2). Therefore, it is significantly slower.

    This algorithm is implemented as a reference and for comparison purpose, but not
    used in the actual GME algorithm.

    """
    n = tree[0, 4] * 2 - 1
    taxas = {}
    for node in postodr[:n]:
        left, right = tree[node, :2]
        if not left:
            taxas[node] = frozenset([right])
        else:
            taxas[node] = taxas[left] | taxas[right]

    full = taxas[0] | frozenset([0])

    dk = dm[taxon]
    for node in range(n):
        taxa_lower = taxas[node]
        adk[node, 0] = dk[list(taxa_lower)].mean()
        taxa_upper = full - taxa_lower
        adk[node, 1] = dk[list(taxa_upper)].mean()


def _init_swaps(tree):
    """Initialize branch swapping information.

    It will create three 1-D arrays to store information of all internal branches of
    the tree:

    - `gains`: Overall tree length reduction (larger is better).
    - `sides`: Which side (left: 0, right: 1) of the target node should be swapped.
    - `nodes`: Corresponding node index of the branch.

    Meanwhile, column 7 of the tree array will be filled with the branch index of the
    each node, if it corresponds to one. This column was previous used to store
    postorder index, but that information is not needed during swapping.

    For a tree with n taxa, there should be n - 3 internal branches. The association
    between nodes and internal branches are fixed, no matter how tree is swapped.

    """
    # number of internal branches
    n = tree[0, 4] - 2
    gains = np.zeros((n,), dtype=float)
    sides = np.empty((n,), dtype=int)
    nodes = np.empty((n,), dtype=int)

    # identify all internal branches
    branch = 0
    for node in range(1, tree.shape[0]):
        if tree[node, 0]:
            tree[node, 7] = branch
            nodes[branch] = node
            branch += 1

    return gains, sides, nodes


def _swap_branches_treenode(node1, node2):
    """Swap two nodes in a TreeNode object.

    This function is for testing purpose.

    """
    parent1, parent2 = node1.parent, node2.parent
    parent1.append(node2, False)
    parent2.append(node1, False)


def _swap_branches(target, side, tree, preodr, stack, use_depth=True):
    r"""Swap one child of the target node with its sibling.

                 |                         |
               parent                    parent
               /   \                     /   \
           target  sibling    =>     target  child
            /  \                      /  \
        other  child              other  sibling

    Target must be an internal node. Root and tips are not allowed.

    This function currently doesn't handle postorder, as subsequent algorithms don't
    need this information.

    This function is specifically designed for nearest neighbor interchange (NNI). It
    may be generalized to swapping non-neighbor branches, but that is not currently
    implemented.

    """
    # This function can potentially optimized using Cython. See _insert_taxon.

    child = tree[target, side]
    other = tree[target, 1 - side]  # other child
    parent = tree[target, 2]
    sibling = tree[target, 3]

    c_size = tree[child, 4]
    o_size = tree[other, 4]
    s_size = tree[sibling, 4]

    # update connections
    tree[target, side] = sibling
    tree[target, 3] = child
    tree[target, 4] += s_size - c_size

    p_side = int(tree[parent, 0] == target)  # sibling side of parent
    tree[parent, p_side] = child

    tree[sibling, 2] = target
    tree[sibling, 3] = other

    tree[child, 2] = parent
    tree[child, 3] = target

    tree[other, 3] = sibling

    # locate the clades under the relevant nodes
    c_start = tree[child, 6]
    c_width = c_size * 2 - 1
    c_end = c_start + c_width
    c_clade = preodr[c_start:c_end]

    s_start = tree[sibling, 6]
    s_width = s_size * 2 - 1
    s_end = s_start + s_width
    s_clade = preodr[s_start:s_end]

    o_start = tree[other, 6]
    o_width = o_size * 2 - 1
    o_end = o_start + o_width
    o_clade = preodr[o_start:o_end]

    # update depth (if needed)
    if use_depth:
        tree[c_clade, 5] -= 1
        tree[s_clade, 5] += 1

    # update preorder (naive)
    # _preorder(preodr, tree, stack)
    # tree[:n, 6] = np.argsort(preodr[:n])

    # update postorder (naive)
    # _postorder(postodr, tree, stack)
    # tree[:n, 7] = np.argsort(postodr[:n])

    # update preorder
    # sibling is the left child of parent
    if p_side == 0:
        tree[target, 6] += c_width - s_width
        stack[:c_width] = c_clade
        stack[c_width] = target
        c1_width = c_width + 1

        # sibling_clade, target, child_clade, other_clade =>
        # child_clade, target, sibling_clade, other_clade
        if side == 0:
            tree[s_clade, 6] += c_width + 1
            tree[c_clade, 6] -= s_width + 1
            preodr[(sc1_start := s_start + c1_width) : c_end] = s_clade
            preodr[s_start:sc1_start] = stack[:c1_width]

        # sibling_clade, target, other_clade, child_clade =>
        # child_clade, target, other_clade, sibling_clade
        else:
            tree[s_clade, 6] += c_width + o_width + 1
            tree[c_clade, 6] -= s_width + o_width + 1
            tree[o_clade, 6] += c_width - s_width
            stack[c1_width : (c1o_width := c1_width + o_width)] = o_clade
            preodr[(sc1o_start := s_start + c1o_width) : c_end] = s_clade
            preodr[s_start:sc1o_start] = stack[:c1o_width]

    # sibling is the right child of parent
    else:
        stack[:s_width] = s_clade

        # target, child_clade, other_clade, sibling_clade =>
        # target, sibling_clade, other_clade, child_clade
        if side == 0:
            tree[c_clade, 6] += (so_width := s_width + o_width)
            tree[s_clade, 6] -= c_width + o_width
            tree[o_clade, 6] += s_width - c_width
            stack[s_width:so_width] = o_clade
            preodr[(cso_start := c_start + so_width) : s_end] = c_clade
            preodr[c_start:cso_start] = stack[:so_width]

        # target, other_clade, child_clade, sibling_clade =>
        # target, other_clade, sibling_clade, child_clade
        else:
            tree[c_clade, 6] += s_width
            tree[s_clade, 6] -= c_width
            preodr[(cs_start := c_start + s_width) : s_end] = c_clade
            preodr[c_start:cs_start] = stack[:s_width]



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main

import numpy as np
import numpy.testing as npt

from skbio import TreeNode
from skbio.tree._compare import rf_dists, wrf_dists, path_dists


class CompareTests(TestCase):

    def test_rf_dists(self):
        """Calculate Robinson-Foulds distances among trees."""
        # all trees have identical taxa
        nwks = ["(((a,b),c),d,e);",
                "((a,(b,c)),d,e);",
                "((a,b),(c,d),e);",
                "(a,b,(c,(d,e)));"]
        trees = [TreeNode.read([x]) for x in nwks]

        obs = rf_dists(trees)
        self.assertSequenceEqual(obs.ids, list("0123"))
        exp = np.array([[0, 2, 2, 0],
                        [2, 0, 4, 2],
                        [2, 4, 0, 2],
                        [0, 2, 2, 0]], dtype=float)
        npt.assert_array_almost_equal(obs.data, exp)

        # with IDs
        ids = list("ABCD")
        obs = rf_dists(trees, ids)
        self.assertSequenceEqual(obs.ids, ids)
        npt.assert_array_almost_equal(obs.data, exp)

        # rooted
        obs = rf_dists(trees, rooted=True).data
        exp = np.array([[0, 2, 2, 4],
                        [2, 0, 4, 4],
                        [2, 4, 0, 4],
                        [4, 4, 4, 0]], dtype=float)
        npt.assert_array_almost_equal(obs, exp)

        # proportion
        obs = rf_dists(trees, proportion=True).data
        exp = np.array([[0, .5, .5, 0],
                        [.5, 0, 1, .5],
                        [.5, 1, 0, .5],
                        [0, .5, .5, 0]], dtype=float)
        npt.assert_array_almost_equal(obs, exp)

        # with unique taxa
        nwks = ["(((a,b),c),((d,e),f),g);",
                "((a,(b,c)),(d,(e,f)),h);",
                "(((a,b),(c,e)),(f,h),i);",
                "((a,((b,c),f)),g,(h,e));"]
        trees = [TreeNode.read([x]) for x in nwks]
        obs = rf_dists(trees).data
        exp = np.array([[0, 2, 2, 4],
                        [2, 0, 4, 2],
                        [2, 4, 0, 4],
                        [4, 2, 4, 0]], dtype=float)
        npt.assert_array_almost_equal(obs, exp)

        # pairwise sharing
        obs = rf_dists(trees, shared_by_all=False).data
        exp = np.array([[0, 4, 2, 6],
                        [4, 0, 6, 4],
                        [2, 6, 0, 6],
                        [6, 4, 6, 0]], dtype=float)
        npt.assert_array_almost_equal(obs, exp)

        # duplicate IDs
        with self.assertRaises(ValueError):
            rf_dists(trees, list("xxxx"))

        # ID number doesn't match
        with self.assertRaises(ValueError):
            rf_dists(trees, list("abc"))

        # no shared taxon
        nwks = ["((a,b),(c,d));",
                "((e,f),(g,h));"]
        trees = [TreeNode.read([x]) for x in nwks]
        obs = rf_dists(trees).data
        self.assertFalse(np.any(obs))

        # only one tree
        obs = rf_dists([TreeNode.read(["((a,b),(c,d));"])]).data
        npt.assert_array_equal(obs, np.zeros((1, 1)))

        # no tree
        with self.assertRaises(ValueError):
            rf_dists([])

    def test_wrf_dists(self):
        """Calculate weighted Robinson-Foulds distances among trees."""
        # all trees have identical taxa
        nwks = ["((a:1,b:2):1,c:4,((d:4,e:5):2,f:6):1);",
                "((a:3,(b:2,c:2):1):3,d:8,(e:5,f:6):2);",
                "((a:1,c:6):2,(b:3,(d:2,e:3):1):2,f:7);"]
        trees = [TreeNode.read([x]) for x in nwks]

        obs = wrf_dists(trees)
        self.assertSequenceEqual(obs.ids, list("012"))
        exp = np.array([[0, 16, 15],
                        [16, 0, 27],
                        [15, 27, 0]], dtype=float)
        npt.assert_array_almost_equal(obs.data, exp)

        # with IDs
        ids = list("ABC")
        obs = wrf_dists(trees, ids)
        self.assertSequenceEqual(obs.ids, ids)
        npt.assert_array_almost_equal(obs.data, exp)

        # rooted
        obs = wrf_dists(trees, rooted=True).data
        exp = np.array([[0, 18, 15],
                        [18, 0, 27],
                        [15, 27, 0]], dtype=float)
        npt.assert_array_almost_equal(obs, exp)

        # Euclidean distances
        obs = wrf_dists(trees, metric="euclidean").data
        exp = np.array([[0, 6.164414, 5],
                        [6.164414, 0, 9.21954446],
                        [5, 9.21954446, 0]])
        npt.assert_array_almost_equal(obs, exp)

        # unit correlation distances
        obs = wrf_dists(trees, metric="unitcorr").data
        exp = np.array([[0, 0.15822237, 0.13211014],
                        [0.15822237, 0, 0.32552865],
                        [0.13211014, 0.32552865, 0]])
        npt.assert_array_almost_equal(obs, exp)

        # invalide distance metric
        with self.assertRaises(ValueError):
            wrf_dists(trees, metric=100)
        with self.assertRaises(AttributeError):
            wrf_dists(trees, metric="hello")

        # with unique taxa
        nwks = ["((a:1,b:2):1,(c:3,d:4):2);",
                "((a:1,(b:2,d:4):1):2,e:5);",
                "(b:2,((c:3,d:4):1,e:5):2);"]
        trees = [TreeNode.read([x]) for x in nwks]
        obs = wrf_dists(trees).data
        exp = np.array([[0, 3, 0],
                        [3, 0, 3],
                        [0, 3, 0]], dtype=float)
        npt.assert_array_almost_equal(obs, exp)

        # pairwise sharing
        obs = wrf_dists(trees, shared_by_all=False).data
        exp = np.array([[0, 4, 0],
                        [4, 0, 6],
                        [0, 6, 0]], dtype=float)
        npt.assert_array_almost_equal(obs, exp)

    def test_path_dists(self):
        """Calculate path-length distances among trees."""
        # all trees have identical taxa
        nwks = ["((a:1,b:2):1,c:4,((d:4,e:5):2,f:6):1);",
                "((a:3,(b:2,c:2):1):3,d:8,(e:5,f:6):2);",
                "((a:1,c:6):2,(b:3,(d:2,e:3):1):2,f:7);"]
        trees = [TreeNode.read([x]) for x in nwks]

        obs = path_dists(trees)
        self.assertSequenceEqual(obs.ids, list("012"))
        exp = np.array([13.7113092, 11.87434209, 19.5192213])
        npt.assert_array_almost_equal(obs.condensed_form(), exp)

        # with IDs
        ids = list("ABC")
        obs = path_dists(trees, ids)
        self.assertSequenceEqual(obs.ids, ids)
        npt.assert_array_almost_equal(obs.condensed_form(), exp)

        # city block distances
        obs = path_dists(trees, metric="cityblock").condensed_form()
        exp = np.array([48, 37, 61])
        npt.assert_array_almost_equal(obs, exp)

        # unit correlation distances
        obs = path_dists(trees, metric="unitcorr").condensed_form()
        exp = np.array([0.14130903, 0.2805951, 0.48826796])
        npt.assert_array_almost_equal(obs, exp)

        # random sampling
        obs = path_dists(trees, sample=3, shuffler=list.sort).condensed_form()
        exp = np.array([4.24264069, 7.87400787, 9.2736185])
        npt.assert_array_almost_equal(obs, exp)

        # too many samples
        with self.assertRaises(ValueError):
            path_dists(trees, sample=10)

        # with unique taxa
        nwks = ["((a:1,b:2):1,(c:3,d:4):2,e:5);",
                "((a:1,(b:2,d:4):1):2,e:5,f:6);",
                "(a:1,b:2,((c:3,d:4):1,e:5):2);"]
        trees = [TreeNode.read([x]) for x in nwks]
        obs = path_dists(trees).condensed_form()
        exp = np.array([4.47213595, 1.73205081, 4.35889894])
        npt.assert_array_almost_equal(obs, exp)

        # pairwise sharing
        obs = path_dists(trees, shared_by_all=False).condensed_form()
        exp = np.array([4.47213595, 2.0, 4.35889894])
        npt.assert_array_almost_equal(obs, exp)


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from warnings import warn, simplefilter


def _warn_once(func, warning, message, param=None):
    r"""Warn once

    Parameters
    ----------
    func : callable
        Function that triggers the warning.
    warning : class
        Warning type.
    message : str
        Warning message.
    param : str, optional
        Parameter of the function that triggers the warning.

    Notes
    -----
    The warning is raised only at the first call of the function. Meanwhile, a "_warned"
    attribute is added to the function, which prevents subsequent warnings.

    If a parameter is given, the warning is raised at the first use of the particular
    parameter, while not affect the host function or other parameters. Meanwhile, a
    "_warned_params" attribute of the function will contain the parameter to prevent
    subsequent warnings.

    """
    if not param:
        if not hasattr(func, "_warned"):
            simplefilter("once", warning)
            warn(message, warning)
            func._warned = True
    else:
        if not hasattr(func, "_warned_params"):
            func._warned_params = set()
        if param not in func._warned_params:
            simplefilter("once", warning)
            warn(message, warning)
            func._warned_params.add(param)


def _deprecation_message(name, ver=None, msg=None, append=True):
    """Create a message indicating deprecation."""
    if msg and not append:
        return msg
    if ver:
        message = f"{name} has been deprecated since {ver}."
    else:
        message = f"{name} is deprecated."
    if msg:
        message += " " + msg
    return message


def _warn_deprecated(func, ver=None, msg=None, append=True):
    r"""Warn of deprecated status of a function.

    Parameters
    ----------
    func : callable
        Function that triggers the warning.
    ver : str, optional
        Version when deprecation became effective.
    msg : str, optional
        A custom warning message.
    append : bool, optional
        Append the custom message to the end of the default message (True, default),
        or replace the entire default message with the custom message (False).

    See Also
    --------
    _warn_param_deprecated

    """
    msg = _deprecation_message(f"`{func.__name__}`", ver, msg, append)
    _warn_once(func, DeprecationWarning, msg)


def _warn_param_deprecated(func, param, ver=None, msg=None, append=True):
    r"""Warn of deprecated status of a parameter of a function.

    Parameters
    ----------
    func : callable
        Function that hosts the parameter.
    param : str
        Parameter that triggers the warning.
    ver : str, optional
    msg : str, optional
    append : bool, optional
        Refer to :func:`_warn_deprecated`.

    See Also
    --------
    _warn_deprecated

    """
    msg = _deprecation_message(
        f"`{func.__name__}`'s parameter `{param}`", ver, msg, append
    )
    _warn_once(func, DeprecationWarning, msg, param)


def _renaming_message(oldname, newname, ver=None):
    """Create a message indicating renaming."""
    if ver:
        message = f"{oldname} was renamed to {newname} in {ver}."
    else:
        message = f"{oldname} has been renamed to {newname}."
    message += " The old name is kept as an alias but is deprecated."
    return message


def _warn_renamed(func, oldname, ver=None):
    """Warn of renaming status of a function."""
    msg = _renaming_message(f"`{oldname}`", f"`{func.__name__}`", ver)
    _warn_once(func, DeprecationWarning, msg)


def _warn_param_renamed(func, param, oldname, ver=None):
    """Warn of renaming status of a parameter of a function."""
    msg = _renaming_message(
        f"`{func.__name__}`'s parameter `{oldname}`", f"`{param}`", ver
    )
    _warn_once(func, DeprecationWarning, msg, param)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from biom import Table, example_table
from skbio.io.registry import Read, Write

Table.default_write_format = "biom"

# Define read and write methods for the Table class.
Table.read = Read()
Table.write = Write()



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
from unittest import TestCase, main
from pathlib import Path
import tempfile

import os
import re
import h5py
import numpy as np

import skbio
from skbio.io import write
from skbio import Protein
from skbio.util import get_data_path
from skbio.embedding._protein import ProteinEmbedding
from skbio.embedding._protein import ProteinVector
from skbio.io.format.embed import (
    _embed_sniffer, _embed_to_generator,
    _generator_to_embed, _embed_to_protein,
    _protein_to_embed, _protein_to_vector,
    _vector_to_protein
)


class TestWriteError(TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()

    def test_write_function(self):
        # note: tiny_embedding_file stores embeddings for our 20 seqs.
        with open(get_data_path('tiny_embedding_file.npz'), 'rb') as fh:
            embeddings = np.load(fh)
            emb_list = [embeddings[arrs] for arrs in embeddings.files]
        # note: pdb_hits.txt stores sequence strings for our 20 seqs.
        seqs = np.loadtxt(get_data_path('pdb_hits.txt'), dtype=str)
        seq_list = [" ".join(list(re.sub(r"[UZOB]", "X", str(seq))))
                    for seq in seqs]

        embed_list = []
        for emb, seq in zip(emb_list, seq_list):
            embed_list.append(ProteinEmbedding(embedding=emb, sequence=seq))

        # create generator object for write testing
        embed_list = (emb for emb in embed_list)
        file_path = os.path.join(self.tempdir.name, 'test_pdb_hits.h5')

        write(embed_list, 'embed', into=get_data_path(file_path))
        self.assertTrue(os.path.exists(file_path))


class EmbedTests(TestCase):
    def setUp(self):
        # single sequence
        rk = 5  # latent dimension of residues
        self.sequences = (
            [
                (
                    np.load(get_data_path('embed1.txt.npy')),
                    Protein(('IGKEEIQQRLAQFVDHWKELKQLAAARGQRL'
                            'EESLEYQQFVANVEEEEAWINEKMTLVASED'),
                            metadata={"id": "seq1"})
                ),
                (
                    np.load(get_data_path('embed2.txt.npy')),
                    Protein(('QQNKELNFKLREKQNEIFELKKIAETLRSKL'
                            'EKYVDITKKLEDQNLNLQIKISDLEKKLSDA'),
                            metadata={"id": "seq2"})
                )
            ]
        )
        self.tempdir = tempfile.TemporaryDirectory()
        tempdir = Path(self.tempdir.name)
        self.writable_emb_path = str(tempdir / Path('test.emb'))
        self.writable_emb_path2 = str(tempdir / Path('test2.emb'))

        self.valid_embed_path = get_data_path('prot.emb')
        self.invalid_embed_path = str(tempdir / Path('invalid'))
        self.nonembed_hdf5_path = str(tempdir / Path('other.hdf5'))

        with open(self.invalid_embed_path, 'wb') as fp:
            fp.write(b'this is not a embed file')

        with h5py.File(self.nonembed_hdf5_path, 'w') as fp:
            fp['stuff'] = [1, 2, 3]

    def test_sniffer(self):
        self.assertEqual(_embed_sniffer(self.valid_embed_path), (True, {}))
        self.assertEqual(_embed_sniffer(self.invalid_embed_path), (False, {}))
        self.assertEqual(_embed_sniffer(self.nonembed_hdf5_path), (False, {}))

    def test_read_write_single(self):
        for emb, seq in self.sequences:
            fh = self.writable_emb_path
            obj = ProteinEmbedding(emb, seq)
            _protein_to_embed(obj, fh)
            emb2 = _embed_to_protein(fh)
            np.testing.assert_array_equal(emb, emb2.embedding)
            self.assertEqual(str(seq), str(emb2))

    def test_read_write_generator(self):
        writable_emb_path2 = 'test2.emb'
        objs1 = [ProteinEmbedding(emb, seq) for emb, seq in self.sequences]
        _generator_to_embed(objs1, self.writable_emb_path2)
        objs2 = _embed_to_generator(self.writable_emb_path2)
        for obj1, obj2 in zip(objs1, objs2):
            np.testing.assert_array_equal(obj1.embedding, obj2.embedding)
            self.assertEqual(str(obj1), str(obj2))

    def test_write_generator(self):
        sequences = [
            (
                np.load(get_data_path('embed1.txt.npy')),
                Protein(('IGKEEIQQRLAQFVDHWKELKQLAAARGQRL'
                        'EESLEYQQFVANVEEEEAWINEKMTLVASED'),
                        metadata={"id": "seq1"})
            ),
            (
                np.load(get_data_path('embed2.txt.npy')),
                Protein(('QQNKELNFKLREKQNEIFELKKIAETLRSKL'
                        'EKYVDITKKLEDQNLNLQIKISDLEKKLSDA'),
                        metadata={"id": "seq2"})
            )
        ]
        f = lambda x: ProteinEmbedding(*x)
        objs1 = (x for x in map(f, sequences))

        tempdir = Path(tempfile.mkdtemp())
        writable_emb_path = str(tempdir / Path('test.emb'))

        skbio.io.write(objs1, format='embed', into=writable_emb_path)
        objs2 = iter(skbio.io.read(writable_emb_path, format='embed',
                                    constructor=ProteinEmbedding))
        for obj1, obj2 in zip(objs1, objs2):
            np.testing.assert_array_equal(obj1.embedding, obj2.embedding)
            self.assertEqual(str(obj1), str(obj2))


class VectorTests(TestCase):
    def setUp(self):
        # single sequence
        rk = 10  # latent dimension of residues
        self.sequences = (
            [
                (
                    np.random.randn(rk),
                    Protein(('IGKEEIQQRLAQFVDHWKELKQLAAARGQRL'
                            'EESLEYQQFVANVEEEEAWINEKMTLVASED'),
                            metadata={"id": "seq1"})
                ),
                (
                    np.random.randn(rk),
                    Protein(('QQNKELNFKLREKQNEIFELKKIAETLRSKL'
                            'EKYVDITKKLEDQNLNLQIKISDLEKKLSDA'),
                            metadata={"id": "seq2"})
                )
            ]
        )
        self.tempdir = tempfile.TemporaryDirectory()
        tempdir = Path(self.tempdir.name)
        self.writable_emb_path = str(tempdir / Path('test.emb'))
        self.writable_emb_path2 = str(tempdir / Path('test2.emb'))

        self.valid_embed_path = get_data_path('prot_vec.emb')
        self.invalid_embed_path = str(tempdir / Path('invalid'))
        self.nonembed_hdf5_path = str(tempdir / Path('other.hdf5'))

        with open(self.invalid_embed_path, 'wb') as fp:
            fp.write(b'this is not a embed file')

        with h5py.File(self.nonembed_hdf5_path, 'w') as fp:
            fp['stuff'] = [1, 2, 3]

    def test_sniffer(self):
        # make sure that the sniffer throws errors as expected
        self.assertEqual(_embed_sniffer(self.valid_embed_path), (True, {}))
        self.assertEqual(_embed_sniffer(self.invalid_embed_path), (False, {}))
        self.assertEqual(_embed_sniffer(self.nonembed_hdf5_path), (False, {}))
        emb, seq = self.sequences[0]
        obj = ProteinVector(emb, seq)
        _protein_to_vector(obj, str(Path(self.tempdir.name) / Path("prot_vec.emb")))

    def test_read_write_single(self):
        for emb, seq in self.sequences:
            fh = self.writable_emb_path
            obj = ProteinVector(emb, seq)
            _protein_to_vector(obj, fh)
            emb2 = _vector_to_protein(fh)
            np.testing.assert_array_equal(
                emb, emb2.embedding.ravel())
            self.assertEqual(str(seq), str(emb2))

    def test_read_write_generator(self):
        writable_emb_path2 = 'test2.emb'
        objs1 = [ProteinVector(emb, seq) for emb, seq in self.sequences]
        _generator_to_embed(objs1, self.writable_emb_path2)
        objs2 = _embed_to_generator(self.writable_emb_path2,
                                    constructor=ProteinVector)
        for obj1, obj2 in zip(objs1, objs2):
            np.testing.assert_array_equal(obj1.embedding, obj2.embedding)
            self.assertEqual(str(obj1), str(obj2))

    def test_write_generator(self):
        sequences = self.sequences
        f = lambda x: ProteinVector(*x)
        objs1 = (x for x in map(f, sequences))

        tempdir = Path(tempfile.mkdtemp())
        writable_emb_path = str(tempdir / Path('test.emb'))

        skbio.io.write(objs1, format='embed', into=writable_emb_path)
        objs2 = iter(skbio.io.read(writable_emb_path, format='embed',
                                    constructor=ProteinVector))
        for obj1, obj2 in zip(objs1, objs2):
            np.testing.assert_array_equal(obj1.embedding, obj2.embedding)
            self.assertEqual(str(obj1), str(obj2))


if __name__ == '__main__':
    main()



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from heapq import heapify

import numpy as np
import numpy.testing as npt

from skbio import DistanceMatrix, TreeNode
from skbio.tree._me import (
    gme,
    bme,
    nni,
    _check_tree,
    _allocate_tree,
    _to_treenode,
    _from_treenode,
    _root_from_treenode,
    _init_tree,
    _insert_taxon_treenode,
    _avgdist_matrix_naive,
    _avgdist_taxon_naive,
    _init_swaps,
    _swap_branches_treenode,
    _swap_branches,
)
from skbio.tree._c_me import (
    _preorder,
    _postorder,
    _insert_taxon,
    _avgdist_matrix,
    _bal_avgdist_matrix,
    _avgdist_taxon,
    _bal_avgdist_taxon,
    _avgdist_d2_insert,
    _bal_avgdist_insert,
    _ols_lengths,
    _ols_lengths_d2,
    _bal_lengths,
    _ols_min_branch_d2,
    _bal_min_branch,
    _avgdist_swap,
    _bal_avgdist_swap,
    _ols_all_swaps,
    _ols_corner_swaps,
    _bal_all_swaps,
)


class MeTests(TestCase):

    def setUp(self):
        # Example 1
        # This newick string was confirmed against http://www.trex.uqam.ca/ which
        # generated the following (isomorphic) newick string:
        # (d:2.0000,e:1.0000,(c:4.0000,(a:2.0000,b:3.0000):3.0000):2.0000);
        #           /-b
        # -a-------|
        #          |          /-c
        #           \--------|
        #                    |          /-e
        #                     \--------|
        #                               \-d
        self.dm1 = np.array([
            [0,  5,  9,  9,  8],
            [5,  0, 10, 10,  9],
            [9, 10,  0,  8,  7],
            [9, 10,  8,  0,  3],
            [8,  9,  7,  3,  0],
        ], dtype=float)
        self.taxa1 = list("abcde")
        self.nwk1 = "(b:3.0,(c:4.0,(e:1.0,d:2.0):2.0):3.0)a:2.0;"
        self.tree1 = np.array([
            [1, 2, 0, 0, 4, 0, 0, 6],  # 0: a (root)
            [0, 1, 0, 2, 1, 1, 1, 0],  # 1: b
            [3, 4, 0, 1, 3, 1, 2, 5],  # 2: (c,(e,d))
            [0, 2, 2, 4, 1, 2, 3, 1],  # 3: c
            [5, 6, 2, 3, 2, 2, 4, 4],  # 4: (e,d)
            [0, 4, 4, 6, 1, 3, 5, 2],  # 5: e
            [0, 3, 4, 5, 1, 3, 6, 3],  # 6: d
        ])
        self.preodr1 = np.array([0, 1, 2, 3, 4, 5, 6])
        self.postodr1 = np.array([1, 3, 5, 6, 4, 2, 0])
        self.lens1 = np.array([2, 3, 3, 4, 2, 1, 2], dtype=float)

        # An unfinished tree for example 1 (missing e)
        #           /-b
        # -a-------|
        #          |          /-c
        #           \--------|
        #                     \-d
        self.nwk1m1 = "(b:3.0,(c:4.0,d:4.0):3.0)a:2.0;"
        self.tree1m1 = np.array([
            [1, 2, 0, 0, 3, 0, 0, 4],  # 0: a (root)
            [0, 1, 0, 2, 1, 1, 1, 0],  # 1: b
            [3, 4, 0, 1, 2, 1, 2, 3],  # 2: (c,d)
            [0, 2, 2, 4, 1, 2, 3, 1],  # 3: c
            [0, 3, 2, 3, 1, 2, 4, 2],  # 4: d
            [0, 0, 0, 0, 0, 0, 0, 0],  # 5: empty
            [0, 0, 0, 0, 0, 0, 0, 0],  # 6: empty
        ])
        self.preodr1m1 = np.array([0, 1, 2, 3, 4, 0, 0])
        self.postodr1m1 = np.array([1, 3, 4, 2, 0, 0, 0])

        # An alternative tree for example 1 (less ladder-like)
        #                     /-b
        #           /--------|
        #          |          \-d
        # -a-------|
        #          |          /-e
        #           \--------|
        #                     \-c
        self.nwk1v2 = "((b,d),(e,c))a;"
        self.tree1v2 = np.array([
            [1, 2, 0, 0, 4, 0, 0, 6],  # 0: a (root)
            [3, 4, 0, 2, 2, 1, 1, 2],  # 1: (b,d)
            [5, 6, 0, 1, 2, 1, 4, 5],  # 2: (e,c)
            [0, 1, 1, 4, 1, 2, 2, 0],  # 3: b
            [0, 3, 1, 3, 1, 2, 3, 1],  # 4: d
            [0, 4, 2, 6, 1, 2, 5, 3],  # 5: e
            [0, 2, 2, 5, 1, 2, 6, 4],  # 6: c
        ])
        self.preodr1v2 = np.array([0, 1, 3, 4, 2, 5, 6])
        self.postodr1v2 = np.array([3, 4, 1, 5, 6, 2, 0])

        # Example 2
        # This example was adopted from the Phylip manual:
        # https://phylipweb.github.io/phylip/doc/neighbor.html
        #           /-Mouse
        # -Bovine--|
        #          |          /-Gibbon
        #           \--------|
        #                    |          /-Orang
        #                     \--------|
        #                              |          /-Gorilla
        #                               \--------|
        #                                        |          /-Human
        #                                         \--------|
        #                                                   \-Chimp
        self.dm2 = np.array([
            [0.0000, 1.6866, 1.7198, 1.6606, 1.5243, 1.6043, 1.5905],
            [1.6866, 0.0000, 1.5232, 1.4841, 1.4465, 1.4389, 1.4629],
            [1.7198, 1.5232, 0.0000, 0.7115, 0.5958, 0.6179, 0.5583],
            [1.6606, 1.4841, 0.7115, 0.0000, 0.4631, 0.5061, 0.4710],
            [1.5243, 1.4465, 0.5958, 0.4631, 0.0000, 0.3484, 0.3083],
            [1.6043, 1.4389, 0.6179, 0.5061, 0.3484, 0.0000, 0.2692],
            [1.5905, 1.4629, 0.5583, 0.4710, 0.3083, 0.2692, 0.0000]
        ])
        self.taxa2 = [
            "Bovine", "Mouse", "Gibbon", "Orang", "Gorilla", "Chimp", "Human"
        ]
        self.nwk2 = (
            "(Mouse:0.76891,(Gibbon:0.372875,(Orang:0.2705917,(Gorilla:0.1525417,"
            "(Human:0.1240375,Chimp:0.1451625):0.0412083):0.03939583):0.0326305)"
            ":0.42026875)Bovine:0.91769;"
        )
        self.tree2 = np.array([
            [ 1,  2,  0,  0,  6,  0,  0, 10],
            [ 0,  1,  0,  2,  1,  1,  1,  0],
            [ 3,  4,  0,  1,  5,  1,  2,  9],
            [ 0,  2,  2,  4,  1,  2,  3,  1],
            [ 5,  6,  2,  3,  4,  2,  4,  8],
            [ 0,  3,  4,  6,  1,  3,  5,  2],
            [ 7,  8,  4,  5,  3,  3,  6,  7],
            [ 0,  4,  6,  8,  1,  4,  7,  3],
            [ 9, 10,  6,  7,  2,  4,  8,  6],
            [ 0,  6,  8, 10,  1,  5,  9,  4],
            [ 0,  5,  8,  9,  1,  5, 10,  5],
        ])
        self.preodr2 = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.postodr2 = np.array([1, 3, 5, 7, 9, 10, 8, 6, 4, 2, 0])
        self.lens2 = np.array([
            0.91769, 0.76891, 0.42026875, 0.372875, 0.0326305, 0.2705917, 0.03939583,
            0.1525417, 0.0412083, 0.1240375, 0.1451625,
        ])

        # Example 3
        # This example uses the same distance matrix as example 2, but with simplified
        # taxon names and different (less ladder-like) tree topology.
        #                               /-d
        #                     /--------|
        #           /--------|          \-c
        #          |         |
        # -a-------|          \-e
        #          |
        #          |          /-b
        #           \--------|
        #                     \-f
        self.dm3 = self.dm2.copy()
        self.taxa3 = list("abcdefg")
        self.nwk3 = "(((d,c),e),(b,f))a;"
        self.tree3 = np.array([
            [1, 2, 0, 0, 5, 0, 0, 8],  # 0, root (a)
            [3, 4, 0, 2, 3, 1, 1, 4],  # 1, ((d,c),e)
            [5, 6, 0, 1, 2, 1, 6, 7],  # 2, (b,f)
            [7, 8, 1, 4, 2, 2, 2, 2],  # 3, (d,c)
            [0, 4, 1, 3, 1, 2, 5, 3],  # 4, e
            [0, 1, 2, 6, 1, 2, 7, 5],  # 5, b
            [0, 5, 2, 5, 1, 2, 8, 6],  # 6, f
            [0, 3, 3, 8, 1, 3, 3, 0],  # 7, d
            [0, 2, 3, 7, 1, 3, 4, 1],  # 8, c
            [0, 0, 0, 0, 0, 0, 0, 0],  # reserved for g's link
            [0, 0, 0, 0, 0, 0, 0, 0],  # reserved for g
        ])
        self.preodr3 = np.array([0, 1, 3, 7, 8, 4, 2, 5, 6, 0, 0])
        self.postodr3 = np.array([7, 8, 3, 4, 1, 5, 6, 2, 0, 0, 0])

        # Example 4
        # This example is based on example 3, with the last taxon g inserted as the
        # sibling of e.
        #                               /-d
        #                     /--------|
        #                    |          \-c
        #           /--------|
        #          |         |          /-e
        #          |          \--------|
        # -a-------|                    \-g
        #          |
        #          |          /-b
        #           \--------|
        #                     \-f
        self.dm4 = self.dm3.copy()
        self.taxa4 = self.taxa3.copy()
        self.nwk4 = "(((d,c),(e,g)),(b,f))a;"
        self.tree4 = np.array([
            [ 1,  2,  0,  0,  6,  0,  0, 10],  # 0, root (a)
            [ 3,  9,  0,  2,  4,  1,  1,  6],  # 1, ((d,c),(e,g))
            [ 5,  6,  0,  1,  2,  1,  8,  9],  # 2, (b,f)
            [ 7,  8,  1,  9,  2,  2,  2,  2],  # 3, (d,c)
            [ 0,  4,  9, 10,  1,  3,  6,  3],  # 4, e
            [ 0,  1,  2,  6,  1,  2,  9,  7],  # 5, b
            [ 0,  5,  2,  5,  1,  2, 10,  8],  # 6, f
            [ 0,  3,  3,  8,  1,  3,  3,  0],  # 7, d
            [ 0,  2,  3,  7,  1,  3,  4,  1],  # 8, c
            [ 4, 10,  1,  3,  2,  2,  5,  5],  # 9, (e,g)
            [ 0,  6,  9,  4,  1,  3,  7,  4],  # 10, g
        ])
        self.preodr4 = np.array([0, 1, 3, 7, 8, 9, 4, 10, 2, 5, 6])
        self.postodr4 = np.array([7, 8, 3, 4, 10, 9, 1, 5, 6, 2, 0])

        # Example 5
        # This dm can yield negative branch lengths.
        self.dm5 = np.array([
            [0,  5,  9,  9,  800],
            [5,  0, 10, 10,  9],
            [9, 10,  0,  8,  7],
            [9, 10,  8,  0,  3],
            [800,  9,  7,  3,  0],
        ], dtype=float)
        self.taxa5 = list("abcde")
        self.nwk5 = "(b:-129,(c:-95,(d:-130,e:133):101):102,a:134);"

    def test_gme(self):
        """The entire tree building workflow."""
        obs = gme(DistanceMatrix(self.dm1, self.taxa1))
        exp = TreeNode.read([self.nwk1])
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)

        obs = gme(DistanceMatrix(self.dm2, self.taxa2))
        exp = TreeNode.read([self.nwk2])
        self.assertEqual(obs.compare_rfd(exp), 0)

        # clip to zero
        dm = DistanceMatrix(self.dm5, self.taxa5)
        obs = gme(dm, neg_as_zero=False)
        exp = TreeNode.read([self.nwk5])
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)
        for taxon in ("b", "c", "d"):
            self.assertTrue(obs.find(taxon).length < 0)
        obs = gme(dm, neg_as_zero=True)
        for taxon in ("b", "c", "d"):
            self.assertEqual(obs.find(taxon).length, 0)

    def test_bme(self):
        """The entire tree building workflow."""
        obs = bme(DistanceMatrix(self.dm1, self.taxa1))
        exp = TreeNode.read([self.nwk1])
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)

        obs = bme(DistanceMatrix(self.dm2, self.taxa2))
        exp = TreeNode.read([self.nwk2])
        self.assertEqual(obs.compare_rfd(exp), 0)

        # clip to zero
        dm = DistanceMatrix(self.dm5, self.taxa5)
        obs = bme(dm, neg_as_zero=False)
        for taxon in ("b", "c", "d"):
            self.assertTrue(obs.find(taxon).length < 0)
        obs = bme(dm, neg_as_zero=True)
        for taxon in ("b", "c", "d"):
            self.assertEqual(obs.find(taxon).length, 0)

    def test_nni(self):
        """The entire tree rearrangement workflow."""
        # Test if NNI can convert an incorrect tree into the groud truth tree.

        # Example 1: In this simple example, FastNNI and BNNI should produce the same
        # topology and branch lengths.
        tree = TreeNode.read([self.nwk1v2])
        tree.append(TreeNode(tree.name, tree.length))
        tree.name, tree.length = None, None

        dm = DistanceMatrix(self.dm1, self.taxa1)

        exp = TreeNode.read([self.nwk1])
        exp.append(TreeNode(exp.name, exp.length))
        exp.name, exp.length = None, None

        obs = nni(tree, dm, balanced=False)
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)

        obs = nni(tree, dm, balanced=True)
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)

        # Example 4: The same, except for branch lengths.
        tree = TreeNode.read([self.nwk4])
        tree.append(TreeNode(tree.name, tree.length))
        tree.name, tree.length = None, None

        dm = DistanceMatrix(self.dm4, self.taxa4)

        exp = TreeNode.read([self.nwk2])
        exp.append(TreeNode(exp.name, exp.length))
        exp.name, exp.length = None, None
        taxonmap = dict(zip(self.taxa2, self.taxa4))
        for tip in exp.tips():
            tip.name = taxonmap[tip.name]

        obs = nni(tree, dm, balanced=False)
        self.assertEqual(obs.compare_rfd(exp), 0)

        obs = nni(tree, dm, balanced=True)
        self.assertEqual(obs.compare_rfd(exp), 0)

        # clip to zero
        dm = DistanceMatrix(self.dm5, self.taxa5)
        tree = TreeNode.read(["(((a,b),d),(c,e));"])
        obs = nni(tree, dm, balanced=False, neg_as_zero=False)
        exp = TreeNode.read([self.nwk5])
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)
        for taxon in ("b", "c", "d"):
            self.assertTrue(obs.find(taxon).length < 0)
        obs = nni(tree, dm, balanced=False, neg_as_zero=True)
        for taxon in ("b", "c", "d"):
            self.assertEqual(obs.find(taxon).length, 0)

    def test_check_tree(self):
        """Check the integrity of a tree structure."""
        _check_tree(self.tree1, self.preodr1, self.postodr1)
        _check_tree(self.tree1m1, self.preodr1m1, self.postodr1m1)
        _check_tree(self.tree1v2, self.preodr1v2, self.postodr1v2)
        _check_tree(self.tree2, self.preodr2, self.postodr2)
        _check_tree(self.tree3, self.preodr3, self.postodr3)
        # TODO: add incorrect examples

    def test_to_treenode(self):
        """Convert an array-based tree structure into a TreeNode object."""
        # entire tree
        obs = _to_treenode(self.tree1, self.taxa1)
        exp = TreeNode.read([self.nwk1])
        self.assertEqual(obs.compare_rfd(exp), 0)

        # with branch lengths
        obs = _to_treenode(self.tree1, self.taxa1, self.lens1)
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)

        # unroot
        obs = _to_treenode(self.tree1, self.taxa1, self.lens1, unroot=True)
        exp = TreeNode.read(["(b:3.0,(c:4.0,(e:1.0,d:2.0):2.0):3.0,a:2.0);"])
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)

        # incomplete tree
        obs = _to_treenode(self.tree1m1, self.taxa1)
        exp = TreeNode.read([self.nwk1m1])
        self.assertEqual(obs.compare_rfd(exp), 0)

        # another example
        obs = _to_treenode(self.tree2, self.taxa2, self.lens2)
        exp = TreeNode.read([self.nwk2])
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_cophenet(exp, ignore_self=True), 0)

    def test_from_treenode(self):
        """Convert a TreeNode object into an array-based tree structure."""
        # a complete tree
        obj = TreeNode.read([self.nwk1])
        taxmap = {x: i for i, x in enumerate(self.taxa1)}
        obs = _allocate_tree(len(taxmap) * 2 - 3)
        _from_treenode(obj, taxmap, *obs)
        _check_tree(*obs)
        npt.assert_array_equal(obs[0], self.tree1)
        npt.assert_array_equal(obs[1], self.preodr1)
        npt.assert_array_equal(obs[2], self.postodr1)

        # an incomplete tree
        obj = TreeNode.read([self.nwk1m1])
        _from_treenode(obj, taxmap, *obs)
        _check_tree(*obs)
        # unused cells are empty, so we need to fill them before comparison
        for arr in obs:
            arr[-2:] = 0
        npt.assert_array_equal(obs[0], self.tree1m1)
        npt.assert_array_equal(obs[1], self.preodr1m1)
        npt.assert_array_equal(obs[2], self.postodr1m1)

        # non-binary tree
        obj = TreeNode.read(["((b,c,d),e)a;"])
        with self.assertRaises(ValueError):
            _from_treenode(obj, taxmap, *obs)

        # another example
        obj = TreeNode.read([self.nwk3])
        taxmap = {x: i for i, x in enumerate(self.taxa3)}
        obs = _allocate_tree(len(taxmap) * 2 - 3)
        _from_treenode(obj, taxmap, *obs)
        _check_tree(*obs)
        # note: self.tree3 is not in preorder so we can't directly compare

    def test_root_from_treenode(self):
        """Convert TreeNode into tree array rooted at 1st taxon."""
        # Example 1: first move the root taxon ("a") to the child place, making the
        # tree trifurcating. This is the typical output of tree-building methods.
        obj = TreeNode.read([self.nwk1])
        obj.append(TreeNode(obj.name))
        obj.name = None

        # Now convert to array to see if it matches the reference.
        obs = _root_from_treenode(obj, self.taxa1)
        _check_tree(*obs)
        npt.assert_array_equal(obs[0], self.tree1)
        npt.assert_array_equal(obs[1], self.preodr1)
        npt.assert_array_equal(obs[2], self.postodr1)

        # Test a different order of taxa. This time, taxon "c" will become the root.
        taxa = list("cadeb")
        obs = _root_from_treenode(obj, taxa)
        exp = (
            np.array([
                [1, 4, 0, 0, 4, 0, 0, 6],  # c
                [2, 3, 0, 4, 2, 1, 1, 2],  # (e,d)
                [0, 3, 1, 3, 1, 2, 2, 0],  # e
                [0, 2, 1, 2, 1, 2, 3, 1],  # d
                [5, 6, 0, 1, 2, 1, 4, 5],  # (b,a)
                [0, 4, 4, 6, 1, 2, 5, 3],  # b
                [0, 1, 4, 5, 1, 2, 6, 4],  # a
            ]),
            np.array([0, 1, 2, 3, 4, 5, 6]),
            np.array([2, 3, 1, 5, 6, 4, 0]),
        )
        _check_tree(*obs)
        for o, e in zip(obs, exp):
            npt.assert_array_equal(o, e)

        # Root the tree at an internal branch and test to recover the same result.
        obj = obj.find("e").parent
        obj = obj.root_at(above=True, branch_attrs=[])
        self.assertEqual(len(obj.children), 2)
        obs = _root_from_treenode(obj, self.taxa1)
        _check_tree(*obs)
        npt.assert_array_equal(obs[0], self.tree1)
        npt.assert_array_equal(obs[1], self.preodr1)
        npt.assert_array_equal(obs[2], self.postodr1)

        # Example 4: Test all possible roots (taxa).
        obj = TreeNode.read([self.nwk4])
        obj.append(TreeNode(obj.name))
        obj.name = None
        taxa = self.taxa4
        m = len(taxa)
        n = m * 2 - 3

        for i in range(m):
            # move target taxon to the first place
            taxon_ = taxa[i]
            taxa_ = [taxon_] + taxa[:i] + taxa[i + 1:]

            # use the algorithm to generate the array
            obs = _root_from_treenode(obj, taxa_)
            _check_tree(*obs)

            # root the TreeNode object instead (slower) and generate the array
            obj_ = obj.root_at(taxon_, branch_attrs=[])
            obj_.prune()

            taxmap = {x: i for i, x in enumerate(taxa_)}
            exp = _allocate_tree(n)
            _from_treenode(obj_, taxmap, *exp)

            for o, e in zip(obs, exp):
                npt.assert_array_equal(o, e)

        # non-binary trees
        taxa = list("abcde")
        nwks = [
            "((a,b,c),(d,e));",  # internal node has 3 children
            "((a),b,(c,(d,e)));",  # internal node has 1 child
            # "(((b,c),((d,e),a)));",  # root has 1 child  # TODO This doesn't work.
            "((a,b),c,d,e);",  # root has 4 children
        ]
        for nwk in nwks:
            obj = TreeNode.read([nwk])
            self.assertRaises(ValueError, _root_from_treenode, obj, taxa)

    def test_preorder(self):
        """Perform preorder traversal."""
        n = self.tree1.shape[0]
        stack = np.full(n, 0)

        # entire tree
        _preorder(obs := np.full(n, 0), self.tree1, stack)
        npt.assert_array_equal(obs, self.preodr1)

        # given clade
        _preorder(obs := np.full(n, 0), self.tree1, stack, start=2)
        exp = np.array([2, 3, 4, 5, 6, 0, 0])
        npt.assert_array_equal(obs, exp)

        # alternative tree
        _preorder(obs := np.full(n, 0), self.tree1v2, stack)
        npt.assert_array_equal(obs, self.preodr1v2)

        # another example
        stack = np.full(n := self.tree3.shape[0], 0)
        _preorder(obs := np.full(n, 0), self.tree3, stack)
        npt.assert_array_equal(obs, self.preodr3)

    def test_postorder(self):
        """Perform postorder traversal."""
        n = self.tree1.shape[0]
        stack = np.full(n, 0)

        # entire tree
        _postorder(obs := np.full(n, 0), self.tree1, stack)
        npt.assert_array_equal(obs, self.postodr1)

        # given clade
        _postorder(obs := np.full(n, 0), self.tree1, stack, start=2)
        exp = np.array([3, 5, 6, 4, 2, 0, 0])
        npt.assert_array_equal(obs, exp)

        # alternative tree
        _postorder(obs := np.full(n, 0), self.tree1v2, stack)
        npt.assert_array_equal(obs, self.postodr1v2)

        # another example
        stack = np.full(n := self.tree3.shape[0], 0)
        _postorder(obs := np.full(n, 0), self.tree3, stack)
        npt.assert_array_equal(obs, self.postodr3)

    def test_allocate_tree(self):
        """Allocate memory space for arrays."""
        n = 5
        tree, preodr, postodr = _allocate_tree(n)
        self.assertTupleEqual(tree.shape, (5, 8))
        self.assertTupleEqual(preodr.shape, (5,))
        self.assertTupleEqual(postodr.shape, (5,))

        # make sure arrays are C-continuous
        for arr in tree, preodr, postodr:
            self.assertTrue(arr.flags.c_contiguous)

    def test_init_tree(self):
        """Initialize triplet tree."""
        dm = np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
        n = 3
        tree, preodr, postodr = _allocate_tree(n)
        ad2 = np.empty((n, 2))
        _init_tree(dm, tree, preodr, postodr, ad2)
        npt.assert_array_equal(tree, np.array([
            [1, 2, 0, 0, 2, 0, 0, 2],
            [0, 1, 0, 2, 1, 1, 1, 0],
            [0, 2, 0, 1, 1, 1, 2, 1],
        ]))
        npt.assert_array_equal(preodr, np.array([0, 1, 2]))
        npt.assert_array_equal(postodr, np.array([1, 2, 0]))
        npt.assert_array_equal(ad2, np.array([
            [0, 0],
            [1, 3],
            [2, 3],
        ]))

        tree, preodr, postodr = _allocate_tree(n)
        adm = np.empty((n, n))
        _init_tree(dm, tree, preodr, postodr, adm, matrix=True)
        np.fill_diagonal(adm, 0)
        npt.assert_array_equal(adm, np.array([
            [0, 1, 2],
            [1, 0, 3],
            [2, 3, 0],
        ]))

    def test_insert_taxon_treenode(self):
        """Insert a taxon between a target node and its parent."""
        # insert into a regular branch
        obs = TreeNode.read([self.nwk1m1])
        _insert_taxon_treenode("e", obs.find("d"), obs)
        exp = TreeNode.read([self.nwk1])
        self.assertEqual(obs.compare_rfd(exp), 0)

        # insert into the root branch
        obs = TreeNode.read([self.nwk1m1])
        _insert_taxon_treenode("e", obs, obs)
        exp = TreeNode.read(["((b,(c,d)),e)a;"])
        self.assertEqual(obs.compare_rfd(exp), 0)

    def test_insert_taxon(self):
        """Insert a taxon between a target node and its parent."""
        # insert taxon e (4) above d (4)
        obs = self.tree1m1.copy(), self.preodr1m1.copy(), self.postodr1m1.copy()
        _insert_taxon(4, 4, *obs)
        exp = np.array([
            [1, 2, 0, 0, 4, 0, 0, 6],
            [0, 1, 0, 2, 1, 1, 1, 0],
            [3, 5, 0, 1, 3, 1, 2, 5],
            [0, 2, 2, 5, 1, 2, 3, 1],
            [0, 3, 5, 6, 1, 3, 5, 2],
            [4, 6, 2, 3, 2, 2, 4, 4],
            [0, 4, 5, 4, 1, 3, 6, 3],
        ]), np.array([
            0, 1, 2, 3, 5, 4, 6
        ]), np.array([
            1, 3, 4, 6, 5, 2, 0
        ])
        for o, e in zip(obs, exp):
            npt.assert_array_equal(o, e)

        # no depth
        obs = self.tree1m1.copy(), self.preodr1m1.copy(), self.postodr1m1.copy()
        _insert_taxon(4, 4, *obs)
        exp[0][:, 5] = [0, 1, 1, 2, 3, 2, 3]
        for o, e in zip(obs, exp):
            npt.assert_array_equal(o, e)

        # insert taxon e (4) above b (1)
        obs = self.tree1m1.copy(), self.preodr1m1.copy(), self.postodr1m1.copy()
        _insert_taxon(4, 1, *obs)
        exp = np.array([
            [5, 2, 0, 0, 4, 0, 0, 6],
            [0, 1, 5, 6, 1, 2, 2, 0],
            [3, 4, 0, 5, 2, 1, 4, 5],
            [0, 2, 2, 4, 1, 2, 5, 3],
            [0, 3, 2, 3, 1, 2, 6, 4],
            [1, 6, 0, 2, 2, 1, 1, 2],
            [0, 4, 5, 1, 1, 2, 3, 1],
        ]), np.array([
            0, 5, 1, 6, 2, 3, 4
        ]), np.array([
            1, 6, 5, 3, 4, 2, 0
        ])
        for o, e in zip(obs, exp):
            npt.assert_array_equal(o, e)

        # insert taxon e (4) above (c,d) (2)
        obs = self.tree1m1.copy(), self.preodr1m1.copy(), self.postodr1m1.copy()
        _insert_taxon(4, 2, *obs)
        exp = np.array([
            [1, 5, 0, 0, 4, 0, 0, 6],
            [0, 1, 0, 5, 1, 1, 1, 0],
            [3, 4, 5, 6, 2, 2, 3, 3],
            [0, 2, 2, 4, 1, 3, 4, 1],
            [0, 3, 2, 3, 1, 3, 5, 2],
            [2, 6, 0, 1, 3, 1, 2, 5],
            [0, 4, 5, 2, 1, 2, 6, 4],
        ]), np.array([
            0, 1, 5, 2, 3, 4, 6
        ]), np.array([
            1, 3, 4, 2, 6, 5, 0
        ])
        for o, e in zip(obs, exp):
            npt.assert_array_equal(o, e)

        # insert taxon e (4) into the root branch (0)
        obs = self.tree1m1.copy(), self.preodr1m1.copy(), self.postodr1m1.copy()
        _insert_taxon(4, 0, *obs)
        exp = np.array([
            [5, 6, 0, 0, 4, 0, 0, 6],
            [0, 1, 5, 2, 1, 2, 2, 0],
            [3, 4, 5, 1, 2, 2, 3, 3],
            [0, 2, 2, 4, 1, 3, 4, 1],
            [0, 3, 2, 3, 1, 3, 5, 2],
            [1, 2, 0, 6, 3, 1, 1, 4],
            [0, 4, 0, 5, 1, 1, 6, 5],
        ]), np.array([
            0, 5, 1, 2, 3, 4, 6
        ]), np.array([
            1, 3, 4, 2, 5, 6, 0
        ])
        for o, e in zip(obs, exp):
            npt.assert_array_equal(o, e)

        # another example: all possible insertions
        tree, taxa = self.tree3, self.taxa3
        m = tree[0, 4] + 1
        n = m * 2 - 3
        taxamap = {}
        for node in self.postodr3:
            if tree[node, 0] == 0:
                taxamap[node] = [taxa[tree[node, 1]]]
            else:
                taxamap[node] = taxamap[tree[node, 0]] + taxamap[tree[node, 1]]
        obj = TreeNode.read([self.nwk3])

        for i in range(n):
            obs = tree.copy(), self.preodr3.copy(), self.postodr3.copy()
            _insert_taxon(m, i, *obs)
            obs = _to_treenode(obs[0], taxa)

            exp = obj.copy()
            _insert_taxon_treenode("g", exp.lca(taxamap[i]), exp)

            self.assertEqual(obs.compare_rfd(exp), 0)

    def test_avgdist_matrix_naive(self):
        """Calculate an average distance matrix using a naive method."""
        # Test if the algorithm reproduces the manually calculated result.
        n = self.tree1.shape[0]
        obs = np.zeros((n, n), dtype=float)
        _avgdist_matrix_naive(obs, self.dm1, self.tree1, self.postodr1)
        exp = np.array([
            [ 0.   ,  5.   ,  8.667,  9.   ,  8.5  ,  8.   ,  9.   ],
            [ 5.   ,  0.   ,  9.667, 10.   ,  9.5  ,  9.   , 10.   ],
            [ 8.667,  9.667,  0.   ,  9.5  ,  9.   ,  8.5  ,  9.5  ],
            [ 9.   , 10.   ,  9.5  ,  0.   ,  7.5  ,  7.   ,  8.   ],
            [ 8.5  ,  9.5  ,  9.   ,  7.5  ,  0.   ,  8.   ,  9.   ],
            [ 8.   ,  9.   ,  8.5  ,  7.   ,  8.   ,  0.   ,  3.   ],
            [ 9.   , 10.   ,  9.5  ,  8.   ,  9.   ,  3.   ,  0.   ],
        ])
        # exp = np.array([
        #     [ 7.75 ,  5.   ,  8.667,  9.   ,  8.5  ,  8.   ,  9.   ],
        #     [ 5.   ,  8.5  ,  9.667, 10.   ,  9.5  ,  9.   , 10.   ],
        #     [ 8.667,  9.667,  9.167,  9.5  ,  9.   ,  8.5  ,  9.5  ],
        #     [ 9.   , 10.   ,  9.5  ,  8.5  ,  7.5  ,  7.   ,  8.   ],
        #     [ 8.5  ,  9.5  ,  9.   ,  7.5  ,  8.5  ,  8.   ,  9.   ],
        #     [ 8.   ,  9.   ,  8.5  ,  7.   ,  8.   ,  6.75 ,  3.   ],
        #     [ 9.   , 10.   ,  9.5  ,  8.   ,  9.   ,  3.   ,  7.5  ],
        # ])
        npt.assert_array_equal(obs.round(3), exp)

        # incomplete tree
        obs = np.zeros((n, n), dtype=float)
        _avgdist_matrix_naive(obs, self.dm1, self.tree1m1, self.postodr1m1)
        exp = np.array([
            [ 0. ,  5. ,  9. ,  9. ,  9. ,  0. ,  0. ],
            [ 5. ,  0. , 10. , 10. , 10. ,  0. ,  0. ],
            [ 9. , 10. ,  0. ,  9.5,  9.5,  0. ,  0. ],
            [ 9. , 10. ,  9.5,  0. ,  8. ,  0. ,  0. ],
            [ 9. , 10. ,  9.5,  8. ,  0. ,  0. ,  0. ],
            [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
            [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
        ])
        npt.assert_array_almost_equal(obs, exp)

        # alternative tree
        obs = np.zeros((n, n), dtype=float)
        _avgdist_matrix_naive(obs, self.dm1, self.tree1v2, self.postodr1v2)
        exp = np.array([
            [ 0.  ,  7.  ,  8.5 ,  5.  ,  9.  ,  8.  ,  9.  ],
            [ 7.  ,  0.  ,  7.5 ,  8.  ,  6.67,  6.  ,  9.  ],
            [ 8.5 ,  7.5 ,  0.  ,  9.5 ,  5.5 ,  6.67,  9.  ],
            [ 5.  ,  8.  ,  9.5 ,  0.  , 10.  ,  9.  , 10.  ],
            [ 9.  ,  6.67,  5.5 , 10.  ,  0.  ,  3.  ,  8.  ],
            [ 8.  ,  6.  ,  6.67,  9.  ,  3.  ,  0.  ,  7.  ],
            [ 9.  ,  9.  ,  9.  , 10.  ,  8.  ,  7.  ,  0.  ],
        ])
        npt.assert_array_equal(obs.round(2), exp)

    def test_avgdist_matrix(self):
        """Calculate an average distance matrix."""
        # Test if the algorithm produces the same result as the native method does.
        dm, tree, preodr, postodr = self.dm1, self.tree1, self.preodr1, self.postodr1
        n = tree.shape[0]
        obs = np.zeros((n, n), dtype=float)
        _avgdist_matrix(obs, dm, tree, preodr, postodr)
        exp = np.zeros((n, n), dtype=float)
        _avgdist_matrix_naive(exp, dm, tree, postodr)
        npt.assert_array_almost_equal(obs, exp)

        # make sure all cells are populated
        _avgdist_matrix(obs := np.full((n, n), np.nan), dm, tree, preodr, postodr)
        np.fill_diagonal(obs, 0)
        _avgdist_matrix_naive(exp := np.zeros((n, n)), dm, tree, postodr)
        npt.assert_array_almost_equal(obs, exp)

        # incomplete tree
        tree, preodr, postodr = self.tree1m1, self.preodr1m1, self.postodr1m1
        _avgdist_matrix(obs := np.zeros((n, n)), dm, tree, preodr, postodr)
        _avgdist_matrix_naive(exp := np.zeros((n, n)), dm, tree, postodr)
        npt.assert_array_almost_equal(obs, exp)

        # example 1 v2
        tree, preodr, postodr = self.tree1v2, self.preodr1v2, self.postodr1v2
        _avgdist_matrix(obs := np.zeros((n, n)), dm, tree, preodr, postodr)
        _avgdist_matrix_naive(exp := np.zeros((n, n)), dm, tree, postodr)
        npt.assert_array_almost_equal(obs, exp)

        # example 2 (complete)
        dm, tree, preodr, postodr = self.dm2, self.tree2, self.preodr2, self.postodr2
        n = tree.shape[0]
        _avgdist_matrix(obs := np.zeros((n, n)), dm, tree, preodr, postodr)
        _avgdist_matrix_naive(exp := np.zeros((n, n)), dm, tree, postodr)
        npt.assert_array_almost_equal(obs, exp)

        # example 3 (incomplete)
        dm, tree, preodr, postodr = self.dm3, self.tree3, self.preodr3, self.postodr3
        n = tree.shape[0]
        _avgdist_matrix(obs := np.zeros((n, n)), dm, tree, preodr, postodr)
        _avgdist_matrix_naive(exp := np.zeros((n, n)), dm, tree, postodr)
        npt.assert_array_almost_equal(obs, exp)

    def test_bal_avgdist_matrix(self):
        """Calculate a balanced average distance matrix."""
        dm, tree, preodr, postodr = self.dm1, self.tree1, self.preodr1, self.postodr1
        n = tree.shape[0]
        _bal_avgdist_matrix(obs := np.zeros((n, n)), dm, tree, preodr, postodr)
        exp = np.array([
            [ 0.  ,  5.  ,  8.75,  9.  ,  8.5 ,  8.  ,  9.  ],
            [ 5.  ,  0.  ,  9.75, 10.  ,  9.5 ,  9.  , 10.  ],
            [ 8.75,  9.75,  0.  ,  9.5 ,  9.  ,  8.5 ,  9.5 ],
            [ 9.  , 10.  ,  9.5 ,  0.  ,  7.5 ,  7.  ,  8.  ],
            [ 8.5 ,  9.5 ,  9.  ,  7.5 ,  0.  ,  7.75,  8.75],
            [ 8.  ,  9.  ,  8.5 ,  7.  ,  7.75,  0.  ,  3.  ],
            [ 9.  , 10.  ,  9.5 ,  8.  ,  8.75,  3.  ,  0.  ],
        ])
        npt.assert_array_almost_equal(obs, exp)

        _bal_avgdist_matrix(obs := np.full((n, n), np.nan), dm, tree, preodr, postodr)
        np.fill_diagonal(obs, 0)
        npt.assert_array_almost_equal(obs, exp)

        dm, tree, preodr, postodr = self.dm3, self.tree3, self.preodr3, self.postodr3
        n = tree[0, 4] * 2 - 1
        _bal_avgdist_matrix(obs := np.zeros((n, n)), dm, tree, preodr, postodr)
        exp = np.array([
            [0.   , 1.607, 1.645, 1.69 , 1.524, 1.687, 1.604, 1.661, 1.72 ],
            [1.607, 0.   , 0.965, 1.362, 1.211, 1.475, 0.455, 1.328, 1.395],
            [1.645, 0.965, 0.   , 1.033, 0.897, 1.581, 1.03 , 0.995, 1.071],
            [1.69 , 1.362, 1.033, 0.   , 0.529, 1.504, 0.562, 0.895, 0.995],
            [1.524, 1.211, 0.897, 0.529, 0.   , 1.446, 0.348, 0.463, 0.596],
            [1.687, 1.475, 1.581, 1.504, 1.446, 0.   , 1.439, 1.484, 1.523],
            [1.604, 0.455, 1.03 , 0.562, 0.348, 1.439, 0.   , 0.506, 0.618],
            [1.661, 1.328, 0.995, 0.895, 0.463, 1.484, 0.506, 0.   , 0.712],
            [1.72 , 1.395, 1.071, 0.995, 0.596, 1.523, 0.618, 0.712, 0.   ],
        ])
        npt.assert_array_equal(obs.round(3), exp)

    def test_avgdist_taxon_naive(self):
        """Calculate taxon-to-subtree average distances using a naive method."""
        # Test if the algorithm reproduces the manually calculated result.
        n = self.tree1.shape[0]
        taxon = self.dm1.shape[0] - 1
        obs = np.zeros((n, 2), dtype=float)
        _avgdist_taxon_naive(obs, taxon, self.dm1, self.tree1m1, self.postodr1m1)
        exp = np.array([
            [6.333, 8.   ],
            [9.   , 6.   ],
            [5.   , 8.5  ],
            [7.   , 6.667],
            [3.   , 8.   ],
            [0.   , 0.   ],
            [0.   , 0.   ],
        ])
        npt.assert_array_equal(obs.round(3), exp)

    def test_avgdist_taxon(self):
        """Calculate taxon-to-subtree average distances."""
        # Test if the algorithm produces the same result as the naive method does.
        dm = self.dm1
        taxon = dm.shape[0] - 1
        tree, preodr, postodr = self.tree1m1, self.preodr1m1, self.postodr1m1
        n = tree.shape[0]
        _avgdist_taxon(obs := np.zeros((n, 2)), taxon, dm, tree, preodr, postodr)
        _avgdist_taxon_naive(exp := np.zeros((n, 2)), taxon, dm, tree, postodr)
        npt.assert_array_almost_equal(obs, exp)

    def test_bal_avgdist_taxon(self):
        """Calculate taxon-to-subtree balanced average distances."""
        dm = self.dm1
        taxon = dm.shape[0] - 1
        tree, preodr, postodr = self.tree1m1, self.preodr1m1, self.postodr1m1
        n = tree.shape[0]
        _bal_avgdist_taxon(obs := np.zeros((n, 2)), taxon, dm, tree, preodr, postodr)
        exp = np.array([
            [7.  , 8.  ],
            [9.  , 6.5 ],
            [5.  , 8.5 ],
            [7.  , 5.75],
            [3.  , 7.75],
            [0.  , 0.  ],
            [0.  , 0.  ],
        ])
        npt.assert_array_almost_equal(obs, exp)

        dm = self.dm3
        taxon = dm.shape[0] - 1
        tree, preodr, postodr = self.tree3, self.preodr3, self.postodr3
        n = tree.shape[0]
        _bal_avgdist_taxon(obs := np.zeros((n, 2)), taxon, dm, tree, preodr, postodr)
        exp = np.array([
            [0.639, 1.59 ],
            [0.411, 1.228],
            [0.866, 1.001],
            [0.515, 0.768],
            [0.308, 0.871],
            [1.463, 0.635],
            [0.269, 1.232],
            [0.471, 0.663],
            [0.558, 0.62 ],
            [0.   , 0.   ],
            [0.   , 0.   ],
        ])
        npt.assert_array_equal(obs.round(3), exp)

    def test_avgdist_d2_insert(self):
        """Update distant-2 subtree average distances after taxon insertion."""
        n = self.tree1.shape[0]
        taxon = self.dm1.shape[0] - 1
        # The following values were taken from the full-scale distance matrix.
        ad2 = np.array([
            [ 0. ,  0. ],  # 0: a (root)
            [ 5. , 10. ],  # 1: b
            [ 9. , 10. ],  # 2: (c,d)
            [ 9.5,  8. ],  # 3: c
            [ 9.5,  8. ],  # 4: d
            [ 0. ,  0. ],  # 5: empty
            [ 0. ,  0. ],  # 6: empty
        ])
        # Insert e as a sibling of d. This should recover tree1.
        target = 4
        _avgdist_taxon(
            adk := np.zeros((n, 2), dtype=float), taxon, self.dm1, self.tree1m1,
            self.preodr1m1, self.postodr1m1
        )
        _avgdist_d2_insert(
            obs := ad2.copy(), target, adk, self.tree1m1, self.preodr1m1
        )
        # The following values were taken from the full-scale distance matrix.
        exp = np.array([
            [0.   , 0.   ],  # 0: a (root)
            [5.   , 9.667],  # 1: b
            [8.667, 9.667],  # 2: (c,(e,d))
            [9.5  , 7.5  ],  # 3: c
            [9.   , 3.   ],  # 4: d
            [9.   , 7.5  ],  # 5: (e,d)
            [8.   , 3.   ],  # 6: e
        ])
        npt.assert_array_equal(obs.round(3), exp)

        # another example: all possible insertions
        dm, tree, preodr, postodr = self.dm3, self.tree3, self.preodr3, self.postodr3
        n = tree.shape[0]
        m = tree[0, 4] + 1
        ran_ = np.arange(n)

        # get distant-2 values from the full matrix
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        ad2 = np.ascontiguousarray(
            np.vstack([adm[ran_, tree[ran_, 2]], adm[ran_, tree[ran_, 3]]]).T)
        _avgdist_taxon(adk := np.zeros((n, 2)), m, dm, tree, preodr, postodr)

        for i in range(n - 2):
            # calculate distant-2 values using the algorithm
            _avgdist_d2_insert(obs := ad2.copy(), i, adk, tree, preodr)

            # insert taxon and calculate full matrix
            tree_, pre_, post_ = tree.copy(), preodr.copy(), postodr.copy()
            _insert_taxon(m, i, tree_, pre_, post_)
            _avgdist_matrix(adm := np.zeros((n, n)), dm, tree_, pre_, post_)
            exp = np.ascontiguousarray(
                np.vstack([adm[ran_, tree_[ran_, 2]], adm[ran_, tree_[ran_, 3]]]).T)

            npt.assert_array_almost_equal(obs, exp)

    def test_bal_avgdist_insert(self):
        """Update balanced average distance matrix after taxon insertion."""
        # Test if the algorithm produces the same result as calculated from the full
        # matrix after taxon insertion.
        tree, preodr, postodr = self.tree1m1, self.preodr1m1, self.postodr1m1
        powers = 2.0 ** -np.arange(5)
        stack = np.zeros(7, dtype=int)
        adm = np.array([
            [ 0. ,  5. ,  9. ,  9. ,  9. ,  0. ,  0. ],
            [ 5. ,  0. , 10. , 10. , 10. ,  0. ,  0. ],
            [ 9. , 10. ,  0. ,  9.5,  9.5,  0. ,  0. ],
            [ 9. , 10. ,  9.5,  0. ,  8. ,  0. ,  0. ],
            [ 9. , 10. ,  9.5,  8. ,  0. ,  0. ,  0. ],
            [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
            [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
        ])
        adk = np.array([
            [7.  , 8.  ],
            [9.  , 6.5 ],
            [5.  , 8.5 ],
            [7.  , 5.75],
            [3.  , 7.75],
            [0.  , 0.  ],
            [0.  , 0.  ],
        ])
        # Insert e as a sibling of d. This should recover tree1.
        target = 4
        _bal_avgdist_insert(
            obs := adm.copy(), target, adk, tree, preodr, postodr, powers, stack
        )
        exp = np.array([
            [ 0.  ,  5.  ,  8.75,  9.  ,  9.  ,  8.5 ,  8.  ],
            [ 5.  ,  0.  ,  9.75, 10.  , 10.  ,  9.5 ,  9.  ],
            [ 8.75,  9.75,  0.  ,  9.5 ,  9.5 ,  9.  ,  8.5 ],
            [ 9.  , 10.  ,  9.5 ,  0.  ,  8.  ,  7.5 ,  7.  ],
            [ 9.  , 10.  ,  9.5 ,  8.  ,  0.  ,  8.75,  3.  ],
            [ 8.5 ,  9.5 ,  9.  ,  7.5 ,  8.75,  0.  ,  7.75],
            [ 8.  ,  9.  ,  8.5 ,  7.  ,  3.  ,  7.75,  0.  ],
        ])
        npt.assert_array_almost_equal(obs, exp)

        # another example: all possible insertions
        dm, tree, preodr, postodr = self.dm3, self.tree3, self.preodr3, self.postodr3
        n = tree.shape[0]
        m = tree[0, 4] + 1
        _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _bal_avgdist_taxon(adk := np.zeros((n, 2)), m, dm, tree, preodr, postodr)
        powers = 2.0 ** -np.arange(m)
        stack = np.zeros(n, dtype=int)

        for i in range(n - 2):
            # update matrix using the algorithm
            _bal_avgdist_insert(
                obs := adm.copy(), i, adk, tree, preodr, postodr, powers, stack
            )

            # insert taxon and calculate full matrix
            tree_, pre_, post_ = tree.copy(), preodr.copy(), postodr.copy()
            _insert_taxon(m, i, tree_, pre_, post_)
            _bal_avgdist_matrix(exp := np.zeros((n, n)), dm, tree_, pre_, post_)

            npt.assert_array_almost_equal(obs, exp)

    def test_ols_lengths(self):
        """Calculate tree branch lengths using an OLS framework."""
        # Example 1: This should recover the same branch lengths as in the original
        # tree.
        dm, tree, preodr, postodr = self.dm1, self.tree1, self.preodr1, self.postodr1
        n = tree.shape[0]
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _ols_lengths(obs := np.zeros(n), adm, tree)
        npt.assert_array_almost_equal(obs, self.lens1)

        # Example 2: The output is close but not precisely identical to those in the
        # original tree.
        dm, tree, preodr, postodr = self.dm2, self.tree2, self.preodr2, self.postodr2
        n = tree.shape[0]
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _ols_lengths(obs := np.zeros(n), adm, tree)
        exp = np.array([
            0.91769, 0.76891, 0.42026875, 0.35793125, 0.04316597, 0.28054444,
            0.03137847, 0.15226875, 0.04148125, 0.12214, 0.14706])
        npt.assert_array_almost_equal(obs, exp)

    def test_ols_lengths_d2(self):
        """Calculate tree branch lengths using an OLS framework."""
        # Test if result matches that calculated from the full matrix.
        # example 1
        dm, tree, preodr, postodr = self.dm1, self.tree1, self.preodr1, self.postodr1
        n = tree.shape[0]
        ran_ = np.arange(n)
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        ad2 = np.ascontiguousarray(
            np.vstack([adm[ran_, tree[ran_, 2]], adm[ran_, tree[ran_, 3]]]).T)
        _ols_lengths_d2(obs := np.zeros(n), ad2, tree)
        _ols_lengths(exp := np.zeros(n), adm, tree)
        npt.assert_array_almost_equal(obs, exp)

        # example 2
        dm, tree, preodr, postodr = self.dm2, self.tree2, self.preodr2, self.postodr2
        n = tree.shape[0]
        ran_ = np.arange(n)
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        ad2 = np.ascontiguousarray(
            np.vstack([adm[ran_, tree[ran_, 2]], adm[ran_, tree[ran_, 3]]]).T)
        _ols_lengths_d2(obs := np.zeros(n), ad2, tree)
        _ols_lengths(exp := np.zeros(n), adm, tree)
        npt.assert_array_almost_equal(obs, exp)

    def test_bal_lengths(self):
        """Calculate tree branch lengths using a balanced framework."""
        # Example 1: In this simple example, OLS and balanced frameworks produce the
        # same branch lengths.
        dm, tree, preodr, postodr = self.dm1, self.tree1, self.preodr1, self.postodr1
        n = tree.shape[0]
        _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _bal_lengths(obs := np.zeros(n), adm, tree)
        npt.assert_array_almost_equal(obs, self.lens1)

        # Example 2: Also slightly different from the original branch lengths.
        dm, tree, preodr, postodr = self.dm2, self.tree2, self.preodr2, self.postodr2
        n = tree.shape[0]
        _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _bal_lengths(obs := np.zeros(n), adm, tree)
        exp = np.array([
            0.92853125, 0.75806875, 0.41086875, 0.36733125, 0.04648125, 0.28469375,
            0.02695625, 0.15393125, 0.03981875, 0.11678125, 0.15241875])
        npt.assert_array_almost_equal(obs, exp)

    def test_ols_min_branch_d2(self):
        """Find the branch with minimum length change using an OLS framework."""
        # Test if result matches ground truth.
        dm = self.dm1
        tree, preodr, postodr = self.tree1m1, self.preodr1m1, self.postodr1m1
        n = tree.shape[0]
        m = tree[0, 4] + 1
        ran_ = np.arange(n)
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        ad2 = np.ascontiguousarray(np.vstack([
            adm[ran_, tree[ran_, 2]], adm[ran_, tree[ran_, 3]]]).T)
        _avgdist_taxon(adk := np.zeros((n, 2)), m, dm, tree, preodr, postodr)
        res = _ols_min_branch_d2(obs := np.zeros(n), ad2, adk, tree, preodr)
        self.assertEqual(res, 4)
        exp = np.array([0, 0, -2, -2, -3, 0, 0], dtype=float)
        # The algorithm omits factor 0.5, therefore we need to x2 here.
        npt.assert_array_almost_equal(obs, 2 * exp)

        # Test if result matches that calculated from the entire tree.
        dm, tree, preodr, postodr = self.dm3, self.tree3, self.preodr3, self.postodr3
        n = tree.shape[0]
        m = tree[0, 4] + 1
        ran_ = np.arange(n)
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        ad2 = np.ascontiguousarray(np.vstack([
            adm[ran_, tree[ran_, 2]], adm[ran_, tree[ran_, 3]]]).T)
        _avgdist_taxon(adk := np.zeros((n, 2)), m, dm, tree, preodr, postodr)
        res = _ols_min_branch_d2(obs := np.zeros(n), ad2, adk, tree, preodr)
        # For each branch, insert taxon, calculate full matrix, then calculate and sum
        # all branch lengths. The difference between each sum and the sum by the root
        # branch is the length change value calculated by the algorithm. 
        exp = np.zeros(n)
        for i in range(n - 2):
            tree_, pre_, post_ = tree.copy(), preodr.copy(), postodr.copy()
            _insert_taxon(m, i, tree_, pre_, post_)
            _avgdist_matrix(adm := np.zeros((n, n)), dm, tree_, pre_, post_)
            _ols_lengths(lens := np.zeros(n), adm, tree_)
            exp[i] = lens.sum()
        exp[:n - 2] -= exp[0]

        npt.assert_array_almost_equal(obs, exp * 2)
        self.assertEqual(res, exp[:n - 2].argmin())

    def test_bal_min_branch(self):
        """Find the branch with minimum length change using a balacned framework."""
        # Test if result matches ground truth.
        dm = self.dm1
        tree, preodr, postodr = self.tree1m1, self.preodr1m1, self.postodr1m1
        n = tree.shape[0]
        m = tree[0, 4] + 1
        _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _bal_avgdist_taxon(adk := np.zeros((n, 2)), m, dm, tree, preodr, postodr)
        res = _bal_min_branch(obs := np.zeros(n), adm, adk, tree, preodr)
        self.assertEqual(res, 4)
        exp = np.array([0, 0, -2, -2, -3, 0, 0], dtype=float)
        # The algorithm omits factor 0.25, therefore we need to x4 here.
        npt.assert_array_almost_equal(obs, 4 * exp)

        # Test if result matches that calculated from the entire tree.
        dm, tree, preodr, postodr = self.dm3, self.tree3, self.preodr3, self.postodr3
        n = tree.shape[0]
        m = tree[0, 4] + 1
        _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _bal_avgdist_taxon(adk := np.zeros((n, 2)), m, dm, tree, preodr, postodr)
        res = _bal_min_branch(obs := np.zeros(n), adm, adk, tree, preodr)
        exp = np.zeros(n)
        for i in range(n - 2):
            tree_, pre_, post_ = tree.copy(), preodr.copy(), postodr.copy()
            _insert_taxon(m, i, tree_, pre_, post_)
            _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree_, pre_, post_)
            _bal_lengths(lens := np.zeros(n), adm, tree_)
            exp[i] = lens.sum()
        exp[:n - 2] -= exp[0]

        npt.assert_array_almost_equal(obs, 4 * exp)
        self.assertEqual(res, exp[:n - 2].argmin())

    def test_swap_branches(self):
        # example 1: swap (e,d) with b
        tree, preodr, postodr = self.tree1, self.preodr1, self.postodr1
        _swap_branches(2, 1, tree, preodr, stack := np.full(tree.shape[0], 0))

        # because _swap_branches doesn't handle postorder yet, weed need to manually
        # reconstruct this information
        _postorder(postodr, tree, stack)
        tree[:, 7] = np.argsort(postodr)

        # make sure tree structure is valid
        _check_tree(tree, preodr, postodr)

        # check new parent mapping
        self.assertEqual(tree[4, 2], 0)
        self.assertEqual(tree[1, 2], 2)

        # example 4: all possible swaps
        tree, preodr, postodr = self.tree4, self.preodr4, self.postodr4
        taxa = self.taxa4
        n = tree.shape[0]
        stack = np.full(n, 0)

        taxamap = {}
        for node in postodr:
            if tree[node, 0] == 0:
                taxamap[node] = [taxa[tree[node, 1]]]
            else:
                taxamap[node] = taxamap[tree[node, 0]] + taxamap[tree[node, 1]]

        for node in range(1, n):
            if tree[node, 0] == 0:
                continue
            node1 = taxamap[tree[node, 3]]
            for side in range(2):
                tree_, pre_, post_ = tree.copy(), preodr.copy(), postodr.copy()
                node2 = taxamap[tree[node, side]]
                exp = _to_treenode(tree_, taxa)
                _swap_branches_treenode(exp.lca(node1), exp.lca(node2))

                _swap_branches(node, side, tree_, pre_, stack)
                _postorder(post_, tree_, stack)
                tree_[:, 7] = np.argsort(post_)
                _check_tree(tree_, pre_, post_)
                obs = _to_treenode(tree_, taxa)

                self.assertEqual(obs.compare_rfd(exp), 0)

    def test_avgdist_swap(self):
        """Update average distance matrix after branch swapping."""
        # Test if the algorithm produces the same result as calculated from the full
        # matrix after branch swapping.

        # example 1: swap (e,d) with b
        dm, tree, preodr, postodr = self.dm1, self.tree1, self.preodr1, self.postodr1
        n = tree.shape[0]
        stack = np.full(n, 0)
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _swap_branches(2, 1, tree, preodr, stack)
        _avgdist_swap(obs := adm.copy(), 2, 1, tree)
        _postorder(postodr, tree, stack)
        tree[:, 7] = np.argsort(postodr)
        _avgdist_matrix(exp := np.zeros((n, n)), dm, tree, preodr, postodr)
        npt.assert_array_almost_equal(obs, exp)

        # example 4: all possible swaps
        dm, tree, preodr, postodr = self.dm4, self.tree4, self.preodr4, self.postodr4
        n = tree.shape[0]
        stack = np.full(n, 0)
        _avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        for node in range(1, n):
            if tree[node, 0] == 0:
                continue
            for side in range(2):
                tree_, pre_, post_ = tree.copy(), preodr.copy(), postodr.copy()
                _swap_branches(node, side, tree_, pre_, stack)
                _avgdist_swap(obs := adm.copy(), node, side, tree_)
                _postorder(post_, tree_, stack)
                tree_[:, 7] = np.argsort(post_)
                _avgdist_matrix(exp := np.zeros((n, n)), dm, tree_, pre_, post_)
                npt.assert_array_almost_equal(obs, exp)

    def test_bal_avgdist_swap(self):
        """Update balanced average distance matrix after branch swapping."""
        # Test if the algorithm produces the same result as calculated from the full
        # matrix after branch swapping.

        # example 1: swap (e,d) with b
        dm, tree, preodr, postodr = self.dm1, self.tree1, self.preodr1, self.postodr1
        n = tree.shape[0]
        m = tree[0, 4] + 1
        stack = np.full(n, 0)
        powers = np.ldexp(1.0, -np.arange(m))
        _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        _swap_branches(2, 1, tree, preodr, stack)
        _bal_avgdist_swap(obs := adm.copy(), 2, 1, tree, preodr, powers, stack)
        _postorder(postodr, tree, stack)
        tree[:, 7] = np.argsort(postodr)
        _bal_avgdist_matrix(exp := np.zeros((n, n)), dm, tree, preodr, postodr)
        npt.assert_array_almost_equal(obs, exp)

        # example 4: all possible swaps
        dm, tree, preodr, postodr = self.dm4, self.tree4, self.preodr4, self.postodr4
        n = tree.shape[0]
        m = tree[0, 4] + 1
        stack = np.full(n, 0)
        powers = np.ldexp(1.0, -np.arange(m))
        _bal_avgdist_matrix(adm := np.zeros((n, n)), dm, tree, preodr, postodr)
        for node in range(1, n):
            if tree[node, 0] == 0:
                continue
            for side in range(2):
                tree_, pre_, post_ = tree.copy(), preodr.copy(), postodr.copy()
                _swap_branches(node, side, tree_, pre_, stack)
                _bal_avgdist_swap(
                    obs := adm.copy(), node, side, tree_, pre_, powers, stack
                )
                _postorder(post_, tree_, stack)
                tree_[:, 7] = np.argsort(post_)
                _bal_avgdist_matrix(exp := np.zeros((n, n)), dm, tree_, pre_, post_)
                npt.assert_array_almost_equal(obs, exp)

    def test_init_swaps(self):
        """Initialize branch swapping information."""
        # There are four internal branches in the tree:
        # 1: ((d,c),(e,g)), 2: (b,f), 3: (d,c), 9: (e,g)
        tree = self.tree4
        _, _, nodes = _init_swaps(tree)
        npt.assert_array_equal(nodes, np.array([1, 2, 3, 9]))

    def test_ols_all_swaps(self):
        """Evaluate possible swaps at all branches of a tree."""
        dm, tree, preodr, postodr = self.dm4, self.tree4, self.preodr4, self.postodr4
        n = tree.shape[0]
        _avgdist_matrix(adm := np.empty((n, n)), dm, tree, preodr, postodr)
        _ols_lengths(lens := np.empty(n), adm, tree)
        lensum = lens.sum()

        # Nodes 2 and 3 will have length reduction.
        _ols_all_swaps(lens, tree, adm)
        exp = np.array([0, 0, -0.47035, -0.03426528, 0, 0, 0, 0, 0, 0, 0])
        npt.assert_array_almost_equal(lens, exp)

        # For each of those two branches, perform branch swapping, update average
        # distance matrix, calculate total tree length, and check if the actual
        # tree length gain equals to the algorithm-calculated value.
        stack = np.full(n, 0)
        for node in np.nonzero(lens)[0]:
            side = tree[node, 7]
            tree_, pre_ = tree.copy(), preodr.copy()
            _swap_branches(node, side, tree_, pre_, stack, use_depth=False)
            _avgdist_swap(adm_ := adm.copy(), node, side, tree_)
            _ols_lengths(lens_ := np.empty(n), adm_, tree_)

            # The algorithm omits factor 0.5, therefore we need to x2 here.
            self.assertAlmostEqual(lens[node], 2 * (lens_.sum() - lensum))

    def test_ols_corner_swaps(self):
        """Update swaps of the four corner branches of a swapped branch."""
        dm, tree, preodr, postodr = self.dm4, self.tree4, self.preodr4, self.postodr4
        n = tree.shape[0]
        _avgdist_matrix(adm := np.empty((n, n)), dm, tree, preodr, postodr)
        _ols_all_swaps(lens := np.empty(n), tree, adm)
        heap = [(lens[i], i, tree[i, 7]) for i in np.nonzero(lens)[0]]
        heapify(heap)

        # Perform swaps at the two branches that are known to produce gains (see
        # `test_ols_all_swaps`), and check if the updated result matches the result
        # re-calculated from the entire tree.
        stack = np.full(n, 0)
        for node in np.nonzero(lens)[0]:
            side = tree[node, 7]
            tree_, pre_ = tree.copy(), preodr.copy()
            _swap_branches(node, side, tree_, pre_, stack, use_depth=False)
            _avgdist_swap(adm_ := adm.copy(), node, side, tree_)
            _ols_all_swaps(exp := np.empty(n), tree_, adm_)
            _ols_corner_swaps(node, heap, obs := lens.copy(), tree_, adm_)
            obs[node] = 0
            npt.assert_array_almost_equal(obs, exp)

    def test_bal_all_swaps(self):
        """Evaluate possible swaps at all branches of a tree."""
        # Using the same strategy as `test_ols_all_swaps`.
        dm, tree, preodr, postodr = self.dm4, self.tree4, self.preodr4, self.postodr4
        n = tree.shape[0]
        m = tree[0, 4] + 1
        powers = np.ldexp(1.0, -np.arange(m))
        _bal_avgdist_matrix(adm := np.empty((n, n)), dm, tree, preodr, postodr)
        _bal_lengths(lens := np.empty(n), adm, tree)
        lensum = lens.sum()

        gains, sides, nodes = _init_swaps(tree)
        _bal_all_swaps(gains, sides, nodes, adm, tree)
        exp = np.array([0, 0.9407, 0.06885, 0])
        npt.assert_array_almost_equal(gains, exp)

        stack = np.full(n, 0)
        for branch in range(nodes.shape[0]):
            if gains[branch] == 0:
                continue
            node = nodes[branch]
            side = sides[branch]
            tree_, pre_ = tree.copy(), preodr.copy()
            _swap_branches(node, side, tree_, pre_, stack, use_depth=True)
            _bal_avgdist_swap(adm_ := adm.copy(), node, side, tree_, pre_, powers, stack)
            _bal_lengths(lens_ := np.empty(n), adm_, tree_)

            # The algorithm omits factor 0.25, therefore we need to x4 here.
            self.assertAlmostEqual(gains[branch], 4 * (lensum - lens_.sum()))


if __name__ == "__main__":
    main()
